# -*- coding: utf-8 -*-
"""Test methods related to reparameterisations"""
import numpy as np
from nessai.livepoint import get_dtype, numpy_array_to_live_points
from nessai.model import Model
from nessai.proposal import FlowProposal
from nessai.reparameterisations import (
    NullReparameterisation,
    RescaleToBounds,
    get_reparameterisation,
)
import pytest
from unittest.mock import MagicMock, Mock, call, patch


@pytest.fixture
def proposal(proposal):
    """Specific mocked proposal for reparameterisation tests"""
    proposal.use_default_reparameterisations = False
    proposal.reverse_reparameterisations = False
    proposal.model = MagicMock(spec=Model)
    return proposal


@pytest.fixture
def dummy_rc():
    """Dummy reparameteristation class"""
    m = MagicMock()
    m.__name__ = "DummyReparameterisation"
    return m


@pytest.fixture
def dummy_cmb_rc():
    """Dummy combined reparameteristation class"""
    m = MagicMock()
    m.add_reparameterisation = MagicMock()
    return m


def test_default_reparameterisation(proposal):
    """Test to make sure default reparameterisation does not cause errors
    for default proposal.
    """
    FlowProposal.add_default_reparameterisations(proposal)


@patch("nessai.proposal.flowproposal.get_reparameterisation")
def test_get_reparamaterisation(mocked_fn, proposal):
    """Make sure the underlying function is called"""
    FlowProposal.get_reparameterisation(proposal, "angle")
    mocked_fn.assert_called_once_with("angle")


@pytest.mark.parametrize("reverse_order", [False, True])
@pytest.mark.parametrize("use_default_reparameterisations", [False, True])
def test_configure_reparameterisations_dict(
    proposal,
    dummy_cmb_rc,
    dummy_rc,
    reverse_order,
    use_default_reparameterisations,
):
    """Test configuration for reparameterisations dictionary.

    Also tests to make sure boundary inversion is set and if the
    `reverse_reparameterisation` is correctly set.
    """
    dummy_rc.return_value = "r"
    # Need to add the parameters before hand to prevent a
    # NullReparameterisation from being added
    dummy_cmb_rc.parameters = ["x"]
    proposal.add_default_reparameterisations = MagicMock()
    proposal.get_reparameterisation = MagicMock(
        return_value=(dummy_rc, {"boundary_inversion": True})
    )
    proposal.model = MagicMock()
    proposal.model.bounds = {"x": [-1, 1], "y": [-1, 1]}
    proposal.model.names = ["x"]
    proposal.reverse_reparameterisations = reverse_order
    proposal.use_default_reparameterisations = use_default_reparameterisations

    with patch(
        "nessai.proposal.flowproposal.CombinedReparameterisation",
        return_value=dummy_cmb_rc,
    ) as mocked_class:
        FlowProposal.configure_reparameterisations(
            proposal, {"x": {"reparameterisation": "default"}}
        )

    proposal.get_reparameterisation.assert_called_once_with("default")

    if use_default_reparameterisations:
        proposal.add_default_reparameterisations.assert_called_once()
    else:
        proposal.add_default_reparameterisations.assert_not_called()

    dummy_rc.assert_called_once_with(
        prior_bounds={"x": [-1, 1]}, parameters="x", boundary_inversion=True
    )
    mocked_class.assert_called_once_with(reverse_order=reverse_order)
    # fmt: off
    proposal._reparameterisation.add_reparameterisations \
        .assert_called_once_with("r")
    # fmt: on

    assert proposal.boundary_inversion is True
    assert proposal.names == ["x"]


@patch("nessai.proposal.flowproposal.CombinedReparameterisation")
def test_configure_reparameterisations_dict_w_params(
    mocked_class, proposal, dummy_rc, dummy_cmb_rc
):
    """Test configuration for reparameterisations dictionary with parameters.

    For example:

        {'x': {'reparmeterisation': 'default', 'parameters': 'y'}}

    This should add both x and y to the reparameterisation.
    """
    dummy_rc.return_value = "r"
    # Need to add the parameters before hand to prevent a
    # NullReparameterisation from being added
    dummy_cmb_rc.parameters = ["x", "y"]
    proposal.add_default_reparameterisations = MagicMock()
    proposal.get_reparameterisation = MagicMock(
        return_value=(
            dummy_rc,
            {},
        )
    )
    proposal.model.bounds = {"x": [-1, 1], "y": [-1, 1]}
    proposal.model.names = ["x", "y"]

    with patch(
        "nessai.proposal.flowproposal.CombinedReparameterisation",
        return_value=dummy_cmb_rc,
    ) as mocked_class:
        FlowProposal.configure_reparameterisations(
            proposal,
            {"x": {"reparameterisation": "default", "parameters": ["y"]}},
        )

    proposal.get_reparameterisation.assert_called_once_with("default")
    proposal.add_default_reparameterisations.assert_not_called()
    dummy_rc.assert_called_once_with(
        prior_bounds={"x": [-1, 1], "y": [-1, 1]},
        parameters=["y", "x"],
    )
    mocked_class.assert_called_once()
    # fmt: off
    proposal._reparameterisation.add_reparameterisations \
        .assert_called_once_with("r")
    # fmt: on

    assert proposal.names == ["x", "y"]


def test_configure_reparameterisations_requires_prime_prior(
    proposal, dummy_rc, dummy_cmb_rc
):
    """
    Test configuration that requires a prime prior but the prime prior is
    missing.
    """
    dummy_rc.return_value = "r"
    # Need to add the parameters before hand to prevent a
    # NullReparameterisation from being added
    dummy_cmb_rc.parameters = ["x", "y"]
    dummy_cmb_rc.has_prime_prior = False
    dummy_cmb_rc.requires_prime_prior = True
    proposal.add_default_reparameterisations = MagicMock()
    proposal.get_reparameterisation = MagicMock(
        return_value=(
            dummy_rc,
            {},
        )
    )
    proposal.model.bounds = {"x": [-1, 1], "y": [-1, 1]}
    proposal.model.names = ["x", "y"]

    with patch(
        "nessai.proposal.flowproposal.CombinedReparameterisation",
        return_value=dummy_cmb_rc,
    ), pytest.raises(RuntimeError) as excinfo:
        FlowProposal.configure_reparameterisations(
            proposal,
            {"x": {"reparameterisation": "default", "parameters": ["y"]}},
        )

    assert "One or more reparameterisations require " in str(excinfo.value)


@patch("nessai.reparameterisations.CombinedReparameterisation")
def test_configure_reparameterisations_dict_missing(mocked_class, proposal):
    """
    Test configuration for reparameterisations dictionary when missing
    the reparameterisation for a parameter.

    This should raise a runtime error.
    """
    proposal.add_default_reparameterisations = MagicMock()
    proposal.get_reparameterisation = get_reparameterisation
    proposal.model.bounds = {"x": [-1, 1], "y": [-1, 1]}
    proposal.model.names = ["x", "y"]

    with pytest.raises(RuntimeError) as excinfo:
        FlowProposal.configure_reparameterisations(
            proposal, {"x": {"scale": 1.0}}
        )

    assert "No reparameterisation found for x" in str(excinfo.value)


def test_configure_reparameterisations_str(proposal):
    """Test configuration for reparameterisations dictionary from a str"""
    proposal.add_default_reparameterisations = MagicMock()
    proposal.get_reparameterisation = get_reparameterisation
    proposal.model.bounds = {"x": [-1, 1], "y": [-1, 1]}
    proposal.model.names = ["x", "y"]
    proposal.fallback_reparameterisation = None
    FlowProposal.configure_reparameterisations(proposal, {"x": "default"})

    proposal.add_default_reparameterisations.assert_not_called()
    assert proposal.rescaled_names == ["x_prime", "y"]
    assert proposal.parameters_to_rescale == ["x"]
    assert proposal._reparameterisation.parameters == ["x", "y"]
    assert proposal._reparameterisation.prime_parameters == ["x_prime", "y"]


def test_configure_reparameterisations_dict_reparam(proposal):
    """Test configuration for reparameterisations dictionary"""
    proposal.add_default_reparameterisations = MagicMock()
    proposal.get_reparameterisation = get_reparameterisation
    proposal.model.bounds = {"x": [-1, 1], "y": [-1, 1]}
    proposal.model.names = ["x", "y"]
    proposal.fallback_reparameterisation = None
    FlowProposal.configure_reparameterisations(
        proposal, {"default": {"parameters": ["x"]}}
    )

    proposal.add_default_reparameterisations.assert_not_called()
    assert proposal.rescaled_names == ["x_prime", "y"]
    assert proposal.parameters_to_rescale == ["x"]
    assert proposal._reparameterisation.parameters == ["x", "y"]
    assert proposal._reparameterisation.prime_parameters == ["x_prime", "y"]


@pytest.mark.parametrize(
    "parameters",
    [
        "x.*",
        [
            "x.*",
        ],
        ("x.*",),
    ],
)
def test_configure_reparameterisations_regex(proposal, parameters):
    """Test configuration for reparameterisations dictionary"""
    proposal.add_default_reparameterisations = MagicMock()
    proposal.get_reparameterisation = get_reparameterisation
    proposal.model.names = ["x_0", "x_1", "y"]
    proposal.model.bounds = {"x_0": [-1, 1], "x_1": [-1, 1], "y": [-1, 1]}
    proposal.fallback_reparameterisation = None
    FlowProposal.configure_reparameterisations(
        proposal,
        {"z-score": {"parameters": parameters}},
    )

    proposal.add_default_reparameterisations.assert_not_called()
    assert proposal.rescaled_names == ["x_0_prime", "x_1_prime", "y"]
    assert proposal.parameters_to_rescale == ["x_0", "x_1"]
    assert proposal._reparameterisation.parameters == ["x_0", "x_1", "y"]
    assert proposal._reparameterisation.prime_parameters == [
        "x_0_prime",
        "x_1_prime",
        "y",
    ]


def test_configure_reparameterisations_none(proposal):
    """Test configuration when input is None"""
    proposal.add_default_reparameterisations = MagicMock()
    proposal.get_reparameterisation = get_reparameterisation
    proposal.model.bounds = {"x": [-1, 1], "y": [-1, 1]}
    proposal.model.names = ["x", "y"]
    proposal.fallback_reparameterisation = None
    FlowProposal.configure_reparameterisations(proposal, None)
    proposal.add_default_reparameterisations.assert_not_called()
    assert proposal.rescaled_names == ["x", "y"]

    assert proposal.parameters_to_rescale == []
    assert proposal._reparameterisation.parameters == ["x", "y"]
    assert proposal._reparameterisation.prime_parameters == ["x", "y"]
    assert all(
        [
            isinstance(r, NullReparameterisation)
            for r in proposal._reparameterisation.reparameterisations.values()
        ]
    )


def test_configure_reparameterisations_fallback(proposal):
    """Test configuration when input is None"""
    proposal.add_default_reparameterisations = MagicMock()
    proposal.get_reparameterisation = get_reparameterisation
    proposal.model.bounds = {"x": [-1, 1], "y": [-1, 1]}
    proposal.model.names = ["x", "y"]
    proposal.fallback_reparameterisation = "default"
    FlowProposal.configure_reparameterisations(proposal, None)
    proposal.add_default_reparameterisations.assert_not_called()
    assert proposal.rescaled_names == ["x_prime", "y_prime"]

    assert proposal.parameters_to_rescale == ["x", "y"]
    assert proposal._reparameterisation.parameters == ["x", "y"]
    assert proposal._reparameterisation.prime_parameters == [
        "x_prime",
        "y_prime",
    ]
    assert all(
        [
            isinstance(r, RescaleToBounds)
            for r in proposal._reparameterisation.reparameterisations.values()
        ]
    )


def test_configure_reparameterisations_incorrect_type(proposal):
    """Assert an error is raised when input is not a dictionary"""
    with pytest.raises(TypeError) as excinfo:
        FlowProposal.configure_reparameterisations(proposal, ["default"])
    assert "must be a dictionary" in str(excinfo.value)


def test_configure_reparameterisations_incorrect_config_type(proposal):
    """Assert an error is raised when the config for a key is not a dictionary
    or a known reparameterisation.
    """
    proposal.model.names = ["x"]
    with pytest.raises(TypeError) as excinfo:
        FlowProposal.configure_reparameterisations(proposal, {"x": ["a"]})
    assert "Unknown config type" in str(excinfo.value)


@pytest.mark.parametrize(
    "reparam",
    [{"z": {"reparameterisation": "sine"}}, {"sine": {"parameters": ["x"]}}],
)
def test_configure_reparameterisation_unknown(proposal, reparam):
    """
    Assert an error is raised if an unknown reparameterisation or parameters
    is passed.
    """
    proposal.model.names = ["x"]
    with pytest.raises(RuntimeError) as excinfo:
        FlowProposal.configure_reparameterisations(proposal, reparam)
    assert "is not a parameter in the model or a known" in str(excinfo.value)


def test_configure_reparameterisation_no_parameters(proposal, dummy_rc):
    """Assert an error is raised if no parameters are specified"""
    proposal.model.names = ["x"]
    proposal.get_reparameterisation = MagicMock(
        return_value=(
            dummy_rc,
            {},
        )
    )
    with pytest.raises(RuntimeError) as excinfo:
        FlowProposal.configure_reparameterisations(
            proposal, {"default": {"update_bounds": True}}
        )
    assert "No parameters key" in str(excinfo.value)


def test_set_rescaling_with_model(proposal, model):
    """
    Test setting the rescaling when the model contains reparmaeterisations.
    """
    proposal.model = model
    proposal.model.reparameterisations = {"x": "default"}

    def update(self):
        proposal.names = model.names
        proposal.parameters_to_rescale = ["x"]
        proposal.rescaled_names = ["x_prime"]

    proposal.configure_reparameterisations = MagicMock()
    proposal.configure_reparameterisations.side_effect = update

    FlowProposal.set_rescaling(proposal)

    proposal.configure_reparameterisations.assert_called_once_with(
        {"x": "default"}
    )
    assert proposal.reparameterisations == {"x": "default"}
    assert proposal.rescaled_names == ["x_prime"]


def test_set_rescaling_with_reparameterisations(proposal, model):
    """
    Test setting the rescaling when a reparameterisations dict is defined.
    """
    proposal.model = model
    proposal.model.reparameterisations = None
    proposal.reparameterisations = {"x": "default"}

    def update(self):
        proposal.names = model.names
        proposal.parameters_to_rescale = ["x"]
        proposal.rescaled_names = ["x_prime"]

    proposal.configure_reparameterisations = MagicMock()
    proposal.configure_reparameterisations.side_effect = update

    FlowProposal.set_rescaling(proposal)

    proposal.configure_reparameterisations.assert_called_once_with(
        {"x": "default"}
    )
    assert proposal.reparameterisations == {"x": "default"}
    assert proposal.rescaled_names == ["x_prime"]


@pytest.mark.parametrize("rescale_parameters", [True, ["x"]])
def test_set_rescaling_parameters(proposal, model, rescale_parameters):
    """Test setting rescaling without reparameterisations."""

    rescale_bounds = [0, 1]
    boundary_inversion = True
    update_bounds = True
    inversion_type = "split"
    detect_edges = False
    detect_edges_kwargs = {}
    rescaled_names = ["x_prime"]

    _rescale_parameters = (
        model.names if rescale_parameters is True else rescale_parameters
    )

    def update(self):
        proposal.names = model.names
        proposal.parameters_to_rescale = _rescale_parameters
        proposal.rescaled_names = rescaled_names

    proposal.model = model
    proposal.model.reparameterisations = None
    proposal.reparameterisations = None
    proposal.configure_reparameterisations = MagicMock()
    proposal.configure_reparameterisations.side_effect = update

    proposal.rescale_parameters = rescale_parameters
    proposal.rescale_bounds = rescale_bounds
    proposal.update_bounds = update_bounds
    proposal.boundary_inversion = boundary_inversion
    proposal.inversion_type = inversion_type
    proposal.detect_edges = detect_edges
    proposal.detect_edges_kwargs = detect_edges_kwargs

    FlowProposal.set_rescaling(proposal)

    reparameterisations = {
        "rescaletobounds": {
            "parameters": _rescale_parameters,
            "rescale_bounds": rescale_bounds,
            "update_bounds": update_bounds,
            "boundary_inversion": boundary_inversion,
            "inversion_type": inversion_type,
            "detect_edges": detect_edges,
            "detect_edges_kwargs": detect_edges_kwargs,
        },
    }

    proposal.configure_reparameterisations.assert_called_with(
        reparameterisations
    )
    proposal.parameters_to_rescale == _rescale_parameters


@pytest.mark.parametrize("n", [1, 10])
def test_rescale(proposal, n):
    """Test rescaling when using reparameterisation dict"""
    x = numpy_array_to_live_points(np.random.randn(n, 2), ["x", "y"])
    x["logL"] = np.random.randn(n)
    x["logP"] = np.random.randn(n)
    x_prime = numpy_array_to_live_points(
        np.random.randn(n, 2), ["x_prime", "y_prime"]
    )
    proposal.x_prime_dtype = get_dtype(["x_prime", "y_prime"])
    proposal._reparameterisation = MagicMock()
    proposal._reparameterisation.reparameterise = MagicMock(
        return_value=[x, x_prime, np.ones(x.size)]
    )

    x_prime_out, log_j = FlowProposal.rescale(
        proposal, x, compute_radius=False, test="lower"
    )

    np.testing.assert_array_equal(
        x_prime[["x_prime", "y_prime"]], x_prime_out[["x_prime", "y_prime"]]
    )
    np.testing.assert_array_equal(
        x[["logP", "logL"]], x_prime_out[["logP", "logL"]]
    )
    proposal._reparameterisation.reparameterise.assert_called_once()


@pytest.mark.parametrize("n", [1, 10])
def test_inverse_rescale(proposal, n):
    """Test rescaling when using reparameterisation dict"""
    x = numpy_array_to_live_points(np.random.randn(n, 2), ["x", "y"]).squeeze()
    x_prime = numpy_array_to_live_points(
        np.random.randn(n, 2), ["x_prime", "y_prime"]
    )
    x_prime["logL"] = np.random.randn(n)
    x_prime["logP"] = np.random.randn(n)
    proposal.x_dtype = get_dtype(["x", "y"])
    proposal._reparameterisation = MagicMock()
    proposal._reparameterisation.inverse_reparameterise = MagicMock(
        return_value=[x, x_prime, np.ones(x.size)]
    )

    x_out, log_j = FlowProposal.inverse_rescale(proposal, x_prime)

    np.testing.assert_array_equal(x[["x", "y"]], x_out[["x", "y"]])
    np.testing.assert_array_equal(
        x_prime[["logP", "logL"]], x_out[["logP", "logL"]]
    )
    proposal._reparameterisation.inverse_reparameterise.assert_called_once()


@pytest.mark.parametrize("has_inversion", [False, True])
def test_verify_rescaling(proposal, has_inversion):
    """Test the method that tests the rescaling at runtime

    Checks both normal parameters and non-sampling parameters (e.g logL)
    """
    x = np.array(
        [(1, np.nan), (2, np.nan)], dtype=[("x", "f8"), ("logL", "f8")]
    )
    x_prime = x["x"] / 2
    log_j = np.array([-2, -2])
    x_out = x.copy()
    log_j_inv = np.array([2, 2])

    if has_inversion:
        x_prime = np.concatenate([x_prime, x_prime])
        log_j = np.concatenate([log_j, log_j])
        log_j_inv = np.concatenate([log_j_inv, log_j_inv])
        x_out = np.concatenate([x_out, x_out])

    proposal.model.new_point = MagicMock(return_value=x)
    proposal.rescale = MagicMock(return_value=(x_prime, log_j))
    proposal.inverse_rescale = MagicMock(return_value=(x_out, log_j_inv))
    proposal.check_state = MagicMock()
    proposal.rescaling_set = True

    FlowProposal.verify_rescaling(proposal)

    proposal.check_state.assert_has_calls(4 * [call(x)])
    # Should call 4 different test cases
    calls = [
        call(x, test="lower"),
        call(x, test="upper"),
        call(x, test=False),
        call(x, test=None),
    ]
    proposal.rescale.assert_has_calls(calls)
    proposal.inverse_rescale.assert_has_calls(4 * [call(x_prime)])


@pytest.mark.parametrize("has_inversion", [False, True])
def test_verify_rescaling_invertible_error(proposal, has_inversion):
    """Assert an error is raised if the rescaling is not invertible"""
    x = np.array([[1], [2]], dtype=[("x", "f8")])
    x_prime = x["x"] / 2
    log_j = np.array([-2, -2])
    x_out = x.copy()[::-1]
    log_j_inv = np.array([2, 2])

    if has_inversion:
        x_prime = np.concatenate([x_prime, x_prime])
        log_j = np.concatenate([log_j, log_j])
        log_j_inv = np.concatenate([log_j_inv, log_j_inv])
        x_out = np.concatenate([x_out, x_out])

    proposal.model.new_point = MagicMock(return_value=x)
    proposal.rescale = MagicMock(return_value=(x_prime, log_j))
    proposal.inverse_rescale = MagicMock(return_value=(x_out, log_j_inv))
    proposal.rescaling_set = True

    with pytest.raises(RuntimeError) as excinfo:
        FlowProposal.verify_rescaling(proposal)
    assert "Rescaling is not invertible for x" in str(excinfo.value)


@pytest.mark.parametrize("has_inversion", [False, True])
def test_verify_rescaling_invertible_error_non_sampling(
    proposal, has_inversion
):
    """Assert an error is raised a non-sampler parameter changes"""
    x = np.array([(1, 3), (2, np.nan)], dtype=[("x", "f8"), ("logL", "f8")])
    x_prime = x["x"] / 2
    log_j = np.array([-2, -2])
    x_out = x.copy()
    log_j_inv = np.array([2, 2])

    if has_inversion:
        x_prime = np.concatenate([x_prime, x_prime])
        log_j = np.concatenate([log_j, log_j])
        log_j_inv = np.concatenate([log_j_inv, log_j_inv])
        x_out = np.concatenate([x_out, x_out])
    # Change the last element, this will test both cases
    x_out["logL"][-1] = 4

    proposal.model.new_point = MagicMock(return_value=x)
    proposal.rescale = MagicMock(return_value=(x_prime, log_j))
    proposal.inverse_rescale = MagicMock(return_value=(x_out, log_j_inv))
    proposal.rescaling_set = True

    with pytest.raises(RuntimeError) as excinfo:
        FlowProposal.verify_rescaling(proposal)
    assert "Non-sampling parameter logL changed" in str(excinfo.value)


@pytest.mark.parametrize("has_inversion", [False, True])
def test_verify_rescaling_jacobian_error(proposal, has_inversion):
    """Assert an error is raised if the Jacobian is not invertible"""
    x = np.array([[1], [2]], dtype=[("x", "f8")])
    x_prime = x["x"] / 2
    log_j = np.array([-2, -2])
    x_out = x.copy()
    log_j_inv = np.array([2, 1])

    if has_inversion:
        x_prime = np.concatenate([x_prime, x_prime])
        log_j = np.concatenate([log_j, log_j])
        log_j_inv = np.concatenate([log_j_inv, log_j_inv])
        x_out = np.concatenate([x_out, x_out])

    proposal.model.new_point = MagicMock(return_value=x)
    proposal.rescale = MagicMock(return_value=(x_prime, log_j))
    proposal.inverse_rescale = MagicMock(return_value=(x_out, log_j_inv))
    proposal.rescaling_set = True

    with pytest.raises(RuntimeError) as excinfo:
        FlowProposal.verify_rescaling(proposal)
    assert "Rescaling Jacobian is not invertible" in str(excinfo.value)


def test_verify_rescaling_rescaling_not_set(proposal):
    """Assert an error is raised if the rescaling is not set"""
    proposal.rescaling_set = False
    with pytest.raises(RuntimeError, match=r"Rescaling must be set .*"):
        FlowProposal.verify_rescaling(proposal)


def test_check_state_update(proposal):
    """Assert the update method is called"""
    x = numpy_array_to_live_points(np.random.randn(10, 2), ["x", "y"])
    proposal._reparameterisation = Mock()
    proposal._reparameterisation.update = MagicMock()
    FlowProposal.check_state(proposal, x)
    proposal._reparameterisation.update.assert_called_once_with(x)



# -*- coding: utf-8 -*-
"""
Main code that handles running and checkpoiting the sampler.
"""
import logging
import os
import signal
import sys
from typing import Optional

from . import config
from .livepoint import live_points_to_dict
from .samplers import NestedSampler, ImportanceNestedSampler
from .posterior import draw_posterior_samples
from .utils import configure_threads
from .utils.torchutils import set_torch_default_dtype
from .utils.io import save_to_json, save_dict_to_hdf5


logger = logging.getLogger(__name__)


class FlowSampler:
    """
    Main class to handle running the nested sampler.

    Parameters
    ----------
    model : :obj:`nessai.model.Model`
        User-defined model.
    output : str, optional
        Output directory
    resume : bool, optional
        If True try to resume the sampler is the resume file exists.
    resume_file : str, optional
        File to resume sampler from.
    resume_data: Any, optional
        Data to resume the sampler from instead of a resume file. The data will
        be passed to the :code:`resume_from_pickled_sampler` of the relevant
        class.
    weights_path : str, optional
        Path to either the weights file or directory containing subdirectories
        with weight files.
    weights_file : str, optional
        Weights file to use, same as :code:`weights_path` and included for
        backwards_compatibility. Will be ignored is :code:`weights_paths` has
        also been specified.
    pytorch_threads : int
        Maximum number of threads to use for torch. If ``None`` torch uses all
        available threads.
    signal_handling : bool
        Enable or disable signal handling.
    exit_code : int, optional
        Exit code to use when forcibly exiting the sampler.
    close_pool : bool
        If True, the multiprocessing pool will be closed once the run method
        has been called. Disables the option in :code:`NestedSampler` if
        enabled.
    disable_vectorisation : bool
        Disable likelihood vectorisation. Overrides the value of
        :py:attr:`nessai.model.Model.allow_vectorised`.
    likelihood_chunksize : Optional[int]
        Chunksize used when evaluating a vectorised likelihood. Overrides the
        of :py:attr:`nessai.model.Model.likelihood_chunksize`. Set to None to
        evaluate the likelihood with all available points.
    allow_multi_valued_likelihood : Optional[bool]
        Allow for a multi-valued likelihood function that will return different
        likelihood values for the same point in parameter space. See
        :py:attr:`nessai.model.Model.allow_multi_valued_likelihood` for more
        details.
    parallelise_prior : Optional[bool]
        If true, and a multiprocessing pool has been specified, then the
        log-prior calculation will be parallelised using multiprocessing.
        If false, then the pool with not be used. Overrides the value of the
        `parallelise_prior` attribute in the model class, which is false by
        default.
    result_extension : str
        Extension used when saving the result format. Defaults to HDF5, but
        also supports JSON.
    kwargs :
        Keyword arguments passed to
        :obj:`~nessai.samplers.nestedsampler.NestedSampler`.
    """

    def __init__(
        self,
        model,
        output=os.getcwd(),
        importance_nested_sampler=False,
        resume=True,
        resume_file="nested_sampler_resume.pkl",
        resume_data=None,
        weights_file=None,
        weights_path=None,
        signal_handling=True,
        exit_code=130,
        pytorch_threads=1,
        close_pool=True,
        eps=None,
        torch_dtype=None,
        disable_vectorisation=False,
        likelihood_chunksize=None,
        allow_multi_valued_likelihood=None,
        parallelise_prior=None,
        result_extension="hdf5",
        **kwargs,
    ):

        configure_threads(
            pytorch_threads=pytorch_threads,
        )

        self._final_samples = None
        self._nested_samples = None
        self.exit_code = exit_code
        self.eps = eps
        if self.eps is not None:
            logger.info(f"Setting eps to {self.eps}")
            config.general.eps = self.eps

        self.torch_dtype = set_torch_default_dtype(torch_dtype)

        if importance_nested_sampler:
            SamplerClass = ImportanceNestedSampler
        else:
            SamplerClass = NestedSampler
        self.importance_nested_sampler = importance_nested_sampler

        self.close_pool = close_pool

        self.result_extension = result_extension

        if disable_vectorisation:
            logger.warning(
                "Overriding value of `allow_vectorised` in the model"
            )
            model.allow_vectorised = False

        if likelihood_chunksize:
            model.likelihood_chunksize = likelihood_chunksize

        if allow_multi_valued_likelihood is not None:
            model.allow_multi_valued_likelihood = allow_multi_valued_likelihood

        if parallelise_prior is not None:
            logger.debug("Overriding `parallelise_prior` in the model")
            model.parallelise_prior = parallelise_prior

        self.output = os.path.join(output, "")
        os.makedirs(self.output, exist_ok=True)
        self.save_kwargs(kwargs)

        model.configure_pool(
            n_pool=kwargs.pop("n_pool", None), pool=kwargs.pop("pool", None)
        )

        if resume and self.check_resume(resume_file, resume_data):
            weights_path = weights_path or weights_file
            if resume_data is not None:
                self.ns = self._resume_from_data(
                    SamplerClass,
                    resume_data=resume_data,
                    model=model,
                    weights_path=weights_path,
                    flow_config=kwargs.get("flow_config"),
                    checkpoint_callback=kwargs.get("checkpoint_callback"),
                )
            else:
                self.ns = self._resume_from_file(
                    SamplerClass,
                    model=model,
                    resume_file=resume_file,
                    weights_path=weights_path,
                    flow_config=kwargs.get("flow_config"),
                    checkpoint_callback=kwargs.get("checkpoint_callback"),
                )
        else:
            logger.debug("Not resuming sampler")
            self.ns = SamplerClass(
                model,
                output=self.output,
                resume_file=resume_file,
                close_pool=False,
                **kwargs,
            )

        if signal_handling:
            try:
                signal.signal(signal.SIGTERM, self.safe_exit)
                signal.signal(signal.SIGINT, self.safe_exit)
                signal.signal(signal.SIGALRM, self.safe_exit)
            except AttributeError:
                logger.error("Cannot set signal attributes on this system")
        else:
            logger.warning(
                "Signal handling is disabled. nessai will not automatically "
                "checkpoint when exited."
            )

    def check_resume(self, resume_file, resume_data):
        """Check if it is possible to resume the sampler"""
        return (
            resume_file
            and any(
                (
                    os.path.exists(os.path.join(self.output, f))
                    for f in [resume_file, resume_file + ".old"]
                )
            )
        ) or resume_data is not None

    def _resume_from_file(
        self,
        SamplerClass,
        resume_file,
        model,
        weights_path,
        flow_config,
        **kwargs,
    ):
        logger.info(f"Trying to resume sampler from {resume_file}")
        try:
            ns = SamplerClass.resume(
                os.path.join(self.output, resume_file),
                model,
                weights_path=weights_path,
                flow_config=flow_config,
                **kwargs,
            )
        except (FileNotFoundError, RuntimeError) as e:
            logger.error(
                f"Could not load resume file from: {self.output} "
                f"with error {e}"
            )
            try:
                resume_file += ".old"
                ns = SamplerClass.resume(
                    os.path.join(self.output, resume_file),
                    model,
                    weights_path=weights_path,
                    flow_config=flow_config,
                    **kwargs,
                )
            except RuntimeError as e:
                logger.error(
                    "Could not load old resume " f"file from: {self.output}"
                )
                raise RuntimeError(
                    "Could not resume sampler " f"with error: {e}"
                )
        return ns

    def _resume_from_data(
        self,
        SamplerClass,
        resume_data,
        model,
        weights_path,
        flow_config,
        **kwargs,
    ):
        logger.info("Trying to resume sampler from `resume_data`")
        return SamplerClass.resume_from_pickled_sampler(
            resume_data,
            model,
            weights_path=weights_path,
            flow_config=flow_config,
            **kwargs,
        )

    @property
    def log_evidence(self):
        """Return the most recent log evidence"""
        return self.logZ

    @property
    def log_evidence_error(self):
        """Return the most recent log evidence error"""
        return self.logZ_error

    @property
    def nested_samples(self):
        """Return the nested samples"""
        if self._final_samples is not None:
            return self._final_samples
        else:
            return self._nested_samples

    def run(
        self,
        plot=True,
        save=True,
        posterior_sampling_method=None,
        close_pool=None,
        **kwargs,
    ):
        """Run the nested sampler.

        Will pick the correct run method given the configuration used.

        Parameters
        ----------
        plot : bool
            Toggle all plots produced once the sampler has converged.
        save : bool, optional
            Toggle automatic saving of results
        posterior_sampling_method : str, optional
            Method used for drawing posterior samples. Defaults to rejection
            sampling.
        close_pool : bool, optional
            Boolean to indicated if the pool should be closed at the end of the
            run function. If False, the user must manually close the pool. If
            specified, this value overrides the value passed when initialising
            the FlowSampler class.
        """
        if self.importance_nested_sampler:
            self.run_importance_nested_sampler(
                plot=plot,
                save=save,
                posterior_sampling_method=posterior_sampling_method,
                close_pool=close_pool,
                **kwargs,
            )
        else:
            self.run_standard_sampler(
                plot=plot,
                save=save,
                posterior_sampling_method=posterior_sampling_method,
                close_pool=close_pool,
                **kwargs,
            )

    def run_standard_sampler(
        self,
        plot=True,
        plot_indices=True,
        plot_posterior=True,
        plot_logXlogL=True,
        save=True,
        posterior_sampling_method=None,
        close_pool=None,
    ):
        """Run the standard nested sampler.

        Parameters
        ----------
        plot : bool
            Toggle all plots produced once the sampler has converged.
        plot_indices : bool
            Toggle the insertion indices plot.
        plot_posterior : bool
            Toggle the posterior distribution plot.
        plot_logXlogL : bool
            Toggle the log-prior volume vs log-likelihood plot.
        save
            Enable or disable saving of a results file.
        posterior_sampling_method
            Method used for drawing posterior samples. Defaults to rejection
            sampling.
        close_pool : bool
            Boolean to indicated if the pool should be closed at the end of the
            run function. If False, the user must manually close the pool. If
            specified, this value overrides the value passed when initialising
            the class.
        """
        if self.importance_nested_sampler:
            raise RuntimeError(
                "Cannot run standard sampler when importance_sampler=True"
            )
        if close_pool is None:
            close_pool = self.close_pool
        if posterior_sampling_method is None:
            posterior_sampling_method = "rejection_sampling"

        self.ns.initialise()
        self.logZ, self._nested_samples = self.ns.nested_sampling_loop()
        self.logZ_error = self.ns.state.log_evidence_error
        logger.info(f"Total sampling time: {self.ns.sampling_time}")
        logger.info(
            "Total likelihood evaluations: "
            f"{self.ns.total_likelihood_evaluations}"
        )

        logger.info("Starting post processing")
        logger.info("Computing posterior samples")
        self.posterior_samples = draw_posterior_samples(
            self._nested_samples,
            log_w=self.ns.state.log_posterior_weights,
            method=posterior_sampling_method,
        )
        logger.info(
            f"Returned {self.posterior_samples.size} " "posterior samples"
        )

        if save:
            self.save_results(
                os.path.join(self.output, "result"),
                extension=self.result_extension,
            )

        if plot:
            from nessai import plot

            if plot_posterior:
                plot.plot_live_points(
                    self.posterior_samples,
                    filename=os.path.join(
                        self.output, "posterior_distribution.png"
                    ),
                )
            if plot_indices:
                plot.plot_indices(
                    self.ns.insertion_indices,
                    self.ns.nlive,
                    filename=os.path.join(
                        self.output, "insertion_indices.png"
                    ),
                )

            if plot_logXlogL:
                self.ns.state.plot(os.path.join(self.output, "logXlogL.png"))

        if close_pool:
            self.ns.close_pool()

    def run_importance_nested_sampler(
        self,
        plot=True,
        plot_posterior=True,
        save=True,
        posterior_sampling_method=None,
        redraw_samples=False,
        n_posterior_samples=None,
        compute_initial_posterior=False,
        close_pool=None,
        **kwargs,
    ):
        """Run the importance nested sampler.

        Parameters
        ----------
        plot
            Enable or disable plotting. Independent of the value passed
            to the :code:`NestedSampler` object.
        save
            Enable or disable saving of a results file.
        posterior_sampling_method
            Method used for drawing posterior samples. Defaults to importance
            sampling.
        redraw_samples
            If True after the sampling is finished, samples are redrawn from
            the meta proposal and used to compute an updated evidence estimate
            and posterior. This can reduce biases in the results.
        n_posterior_samples
            Number of posterior samples to draw when when redrawing samples.
        compute_initial_posterior
            Enables or disables computing the posterior before redrawing
            samples. If :code:`redraw_samples` is False, then this flag is
            ignored.
        close_pool : bool
            Boolean to indicated if the pool should be closed at the end of the
            run function. If False, the user must manually close the pool. If
            specified, this value overrides the value passed when initialising
            the class.
        kwargs
            Keyword arguments passed to \
                :py:meth:`~nessai.importancesampler.ImportanceNestedSampler.draw_final_samples`
        """
        if not self.importance_nested_sampler:
            raise RuntimeError(
                "Cannot run importance sampler when importance_sampler=False"
            )
        if close_pool is None:
            close_pool = self.close_pool
        if posterior_sampling_method is None:
            posterior_sampling_method = "importance_sampling"

        self.ns.nested_sampling_loop()
        self._nested_samples = self.ns.samples
        self.logZ = self.ns.log_evidence
        self.logZ_error = self.ns.log_evidence_error
        logger.info(f"Total sampling time: {self.ns.sampling_time}")
        logger.info(
            "Total likelihood evaluations: "
            f"{self.ns.total_likelihood_evaluations}"
        )

        logger.info("Starting post processing")

        if redraw_samples:
            logger.info("Redrawing samples")
            self.initial_logZ = self.logZ
            self.initial_logZ_error = self.logZ_error
            _, self._final_samples = self.ns.draw_final_samples(
                n_post=n_posterior_samples,
                **kwargs,
            )
            self.logZ = self.ns.final_log_evidence
            self.logZ_error = self.ns.final_log_evidence_error

        logger.info("Computing posterior samples")

        if compute_initial_posterior or not redraw_samples:
            logger.debug("Computing posterior samples")
            self.initial_posterior_samples = self.ns.draw_posterior_samples(
                sampling_method=posterior_sampling_method,
                use_final_samples=False,
            )
        if redraw_samples:
            self.posterior_samples = self.ns.draw_posterior_samples(
                sampling_method=posterior_sampling_method,
                use_final_samples=True,
            )
        else:
            self.posterior_samples = self.initial_posterior_samples
        logger.info(
            f"Returned {self.posterior_samples.size} posterior samples"
        )

        if save:
            self.save_results(
                os.path.join(self.output, "result"),
                extension=self.result_extension,
            )

        if plot and plot_posterior:
            logger.debug("Producing plots")
            from nessai import plot

            plot.plot_live_points(
                self.posterior_samples,
                filename=os.path.join(
                    self.output, "posterior_distribution.png"
                ),
            )
            if redraw_samples and compute_initial_posterior:
                plot.plot_live_points(
                    self.initial_posterior_samples,
                    filename=os.path.join(
                        self.output, "initial_posterior_distribution.png"
                    ),
                )
        if close_pool:
            self.ns.close_pool()

    def save_kwargs(self, kwargs: dict) -> None:
        """
        Save the dictionary of keyword arguments used.

        Uses an encoder class to handle numpy arrays.

        Parameters
        ----------
        kwargs : dict
            Dictionary of kwargs to save.
        """
        d = kwargs.copy()
        d["eps"] = self.eps
        d["torch_dtype"] = self.torch_dtype
        d["importance_sampler"] = self.importance_nested_sampler
        save_to_json(d, os.path.join(self.output, "config.json"))

    def save_results(
        self,
        filename: str,
        extension: Optional[str] = None,
    ) -> None:
        """
        Save the results from sampling to a specific results file.

        Parameters
        ----------
        filename : str
            Name of file to save results to.
        extension : Optional[str]
            File extension to used. If not specified, it will be inferred from
            the filename.
        """
        d = self.ns.get_result_dictionary()

        d["posterior_samples"] = self.posterior_samples
        if hasattr(self, "initial_posterior_samples"):
            d["initial_posterior_samples"] = self.initial_posterior_samples

        ext = os.path.splitext(filename)[1].lstrip(".")
        if extension is None:
            if ext == "":
                raise RuntimeError(
                    "Must specify file extension if not present in filename!"
                )
            else:
                extension = ext
        elif ext == "":
            filename = ".".join([filename, extension])

        if extension == "json":
            d["posterior_samples"] = live_points_to_dict(
                d["posterior_samples"]
            )
            save_to_json(d, filename)
        elif extension in ["hdf5", "h5"]:
            save_dict_to_hdf5(d, filename)
        else:
            raise RuntimeError(f"Unknown file extension: {extension}")

    def terminate_run(self, code=None):
        """Terminate a sampling run.

        Parameters
        ----------
        code : int, optional
            Code passed to :code:`close_pool`
        """
        logger.warning("Terminating run")
        self.ns.close_pool(code=code)
        self.ns.checkpoint()

    def safe_exit(self, signum=None, frame=None):
        """
        Safely exit. This includes closing the multiprocessing pool.
        """
        logger.warning(f"Trying to safely exit with code {signum}")
        self.terminate_run(code=signum)
        logger.warning(f"Exiting with code: {self.exit_code}")
        sys.exit(self.exit_code)



"""
Test the testing utilities
"""

import numpy as np
import pytest

from nessai.utils.testing import (
    IntegrationTestModel,
    assert_structured_arrays_equal,
)


@pytest.mark.parametrize("n", [1, 10])
@pytest.mark.parametrize("dims", [2, 4])
def test_integration_test_model(n, dims):
    """Assert the model is valid"""
    model = IntegrationTestModel(dims)
    model.verify_model()
    x = model.new_point(n)
    log_p = model.log_prior(x)
    log_l = model.log_likelihood(x)
    assert np.isfinite(log_p).all()
    assert np.isfinite(log_l).all()
    assert len(log_p) == len(x)
    assert len(log_l) == len(x)
    x_hyper = model.to_unit_hypercube(x)
    x_re = model.from_unit_hypercube(x_hyper)
    assert_structured_arrays_equal(x_re, x)
    assert len(x_hyper) == len(x)


def test_assert_struct_arrays_different_fields():
    """Assert an errors is raised if the arrays have different fields"""
    x = np.array((1, 2), dtype=[("x", "f8"), ("y", "f8")])
    y = np.array((1, 2), dtype=[("x", "f8"), ("y", "f4")])
    with pytest.raises(AssertionError):
        assert_structured_arrays_equal(x, y)


def test_assert_struct_arrays_equal_values():
    """Assert an errors is raised if fields have different values"""
    dtype = [("x", "f8"), ("y", "f8"), ("z", "f4")]
    x = np.array((1, 2, 3), dtype=dtype)
    y = np.array((1, 3, 4), dtype=dtype)
    with pytest.raises(AssertionError):
        assert_structured_arrays_equal(x, y)


@pytest.mark.parametrize(
    "x, y",
    [
        (
            np.array([(1, 2), (3, 4)], dtype=[("x", "f8"), ("y", "f8")]),
            np.array([(2, 3), (4, 5)], dtype=[("x", "f8"), ("y", "f8")]),
        ),
        (
            np.array([(1, 2), (3, 4)], dtype=[("x", "f8"), ("y", "f8")]),
            np.array(
                [
                    (1, 2),
                ],
                dtype=[("x", "f8"), ("y", "f8")],
            ),
        ),
    ],
)
def test_assert_struct_arrays_equal_array(x, y):
    """Assert an errors is raised with different values with len(x) > 1"""
    with pytest.raises(AssertionError):
        assert_structured_arrays_equal(x, y)


def test_assert_struct_arrays_equal_NaNs():
    """Assert NaNs do not raise an error"""
    dtype = [("x", "f8"), ("y", "f8"), ("z", "f4")]
    x = np.array((1, 2, np.nan), dtype=dtype)
    y = np.array((1, 2, np.nan), dtype=dtype)
    assert_structured_arrays_equal(x, y)


def test_assert_struct_array_equal_tol():
    """Assert arrays that are close do not raised an error if atol is set"""
    dtype = [("x", "f8"), ("y", "f8"), ("z", "f8")]
    x = np.array((1.0, 2.0, 3.0), dtype=dtype)
    y = np.array((1.0, 2.0, 3.0 + 1e-10), dtype=dtype)
    assert_structured_arrays_equal(x, y, atol=1e-9, rtol=0.0)


def test_assert_struct_array_equal_tol_error():
    """Assert arrays that are above the atol raised an error"""
    dtype = [("x", "f8"), ("y", "f8"), ("z", "f8")]
    x = np.array((1.0, 2.0, 3.0), dtype=dtype)
    y = np.array((1.0, 2.0, 3.0 + 1e-10), dtype=dtype)
    with pytest.raises(AssertionError):
        assert_structured_arrays_equal(x, y, atol=1e-11, rtol=0.0)



# -*- coding: utf-8 -*-
"""
Test general config functions called when the nested sampler is initialised.
"""
import os
import numpy as np
import pytest
from unittest.mock import patch, MagicMock

from nessai.samplers.nestedsampler import NestedSampler


@pytest.mark.parametrize("plot", [True, False])
@pytest.mark.parametrize("shrinkage_expectation", ["t", "logt"])
def test_init(sampler, plot, shrinkage_expectation):
    """Test the init method"""
    pool = "pool"
    n_pool = 2
    model = MagicMock()
    model.verify_model = MagicMock()
    model.configure_pool = MagicMock()

    sampler.setup_output = MagicMock()
    sampler.configure_flow_reset = MagicMock()
    sampler.configure_flow_proposal = MagicMock()
    sampler.configure_uninformed_proposal = MagicMock

    with patch("nessai.samplers.nestedsampler._NSIntegralState") as mock_state:
        NestedSampler.__init__(
            sampler,
            model,
            nlive=100,
            poolsize=100,
            pool=pool,
            n_pool=n_pool,
            plot=plot,
            shrinkage_expectation=shrinkage_expectation,
        )
    assert sampler.initialised is False
    assert sampler.nlive == 100
    mock_state.assert_called_once_with(
        100,
        track_gradients=plot,
        expectation=shrinkage_expectation,
    )
    model.verify_model.assert_called_once()
    model.configure_pool.assert_called_once_with(pool=pool, n_pool=n_pool)


@pytest.mark.parametrize("plot", [False, True])
def test_setup_output(sampler, tmpdir, plot):
    """Test setting up the output directories"""
    path = str(tmpdir.mkdir("test"))
    sampler.plot = plot
    rf = "test.pkl"
    with patch(
        "nessai.samplers.base.BaseNestedSampler.configure_output"
    ) as mock:
        NestedSampler.configure_output(sampler, path, rf)
    mock.assert_called_once_with(path, rf)
    if plot:
        assert os.path.exists(os.path.join(path, "diagnostics"))


def test_configure_max_iteration(sampler):
    """Test to make sure the maximum iteration is set correctly"""
    NestedSampler.configure_max_iteration(sampler, 10)
    assert sampler.max_iteration == 10


def test_configure_no_max_iteration(sampler):
    """Test to make sure if no max iteration is given it is set to inf"""
    NestedSampler.configure_max_iteration(sampler, None)
    assert sampler.max_iteration == np.inf


def test_training_frequency(sampler):
    """Make sure training frequency is set"""
    NestedSampler.configure_training_frequency(sampler, 100)
    assert sampler.training_frequency == 100


@pytest.mark.parametrize("f", [None, "inf", "None"])
def test_training_frequency_on_empty(sampler, f):
    """Test the values that should give 'train on empty'"""
    NestedSampler.configure_training_frequency(sampler, f)
    assert sampler.training_frequency == np.inf



#!/usr/bin/env python

"""
Example of parallelising the likelihood evaluation in nessai.

Shows the two methods supported in nessai: setting n_pool or using a
user-defined pool.
"""

import numpy as np
from multiprocessing import Pool

from nessai.flowsampler import FlowSampler
from nessai.model import Model
from nessai.utils import setup_logger
from nessai.utils.multiprocessing import initialise_pool_variables


output = "./outdir/parallelisation_example/"
logger = setup_logger(output=output)


# Generate the data
truth = {"mu": 1.7, "sigma": 0.7}
bounds = {"mu": [-3, 3], "sigma": [0.01, 3]}
n_points = 1000
data = np.random.normal(truth["mu"], truth["sigma"], size=n_points)


class GaussianLikelihood(Model):
    """
    Gaussian likelihood with the mean and standard deviation as the parameters
    to infer.

    Parameters
    ----------
    data : :obj:`numpy.ndarray`
        Array of data.
    bounds : dict
        The prior bounds.
    """

    def __init__(self, data, bounds):
        self.names = list(bounds.keys())
        self.bounds = bounds
        self.data = data

    def log_prior(self, x):
        """Uniform prior on both parameters."""
        log_p = np.log(self.in_bounds(x), dtype="float")
        for bounds in self.bounds.values():
            log_p -= np.log(bounds[1] - bounds[0])
        return log_p

    def log_likelihood(self, x):
        """Gaussian likelihood."""
        log_l = np.sum(
            -np.log(x["sigma"])
            - 0.5 * ((self.data - x["mu"]) / x["sigma"]) ** 2
        )
        return log_l


# Using n_pool
logger.warning("Running nessai with n_pool")
# Configure the sampler with 3 total threads, 2 of which are used for
# evaluating the likelihood.
fs = FlowSampler(
    GaussianLikelihood(data, bounds),
    output=output,
    resume=False,
    seed=1234,
    pytorch_threads=2,  # Allow pytorch to use 2 threads
    n_pool=2,  # Threads for evaluating the likelihood
)

# Run the sampler
fs.run()

# Using a user-defined pool
logger.warning("Running nessai with a user-defined pool")

# Must initialise the global variables for the pool prior to starting it
model = GaussianLikelihood(data, bounds)
initialise_pool_variables(model)
# Define the pool
pool = Pool(2)

fs = FlowSampler(
    model,
    output=output,
    resume=False,
    seed=1234,
    pool=pool,  # User-defined pool
)

# Run the sampler
# The pool will automatically be closed. This can be disabled by passing
# `close_pool=False` to the sampler.
fs.run()



# -*- coding: utf-8 -*-
"""
Utilities related to insertion indices.
"""
import numpy as np
from scipy import stats


def compute_indices_ks_test(indices, nlive, mode="D+"):
    """
    Compute the two-sided KS test for discrete insertion indices for a given
    number of live points

    Parameters
    ----------
    indices : array_like
        Indices of newly inserted live points
    nlive : int
        Number of live points

    Returns
    -------
    D : float
        Two-sided KS statistic
    p : float
        p-value
    """
    if len(indices):
        counts = np.zeros(nlive)
        u, c = np.unique(indices, return_counts=True)
        counts[u] = c
        cdf = np.cumsum(counts) / len(indices)
        if mode == "D+":
            D = np.max(np.arange(1.0, nlive + 1) / nlive - cdf)
        elif mode == "D-":
            D = np.max(cdf - np.arange(0.0, nlive) / nlive)
        else:
            raise RuntimeError(f"{mode} is not a valid mode. Choose D+ or D-")
        p = stats.ksone.sf(D, len(indices))
        return D, p
    else:
        return None, None


def bonferroni_correction(p_values, alpha=0.05):
    """
    Apply the Bonferroni correction for multiple tests.

    Based on the implementation in ``statmodels.stats.multitest``

    Parameters
    ----------
    p_values :  array_like
        Uncorrelated p-values.
    alpha : float, optional
        Family wise error rate.
    """
    p_values = np.asarray(p_values)
    alpha_bon = alpha / p_values.size
    reject = p_values <= alpha_bon
    p_values_corrected = p_values * p_values.size
    p_values_corrected[p_values_corrected > 1] = 1
    return reject, p_values_corrected, alpha_bon



#!/usr/bin/env python

# Example of using the boundary inversion available in nessai.
# See also the reparameterisations example.

import numpy as np
from scipy.stats import norm

from nessai.flowsampler import FlowSampler
from nessai.model import Model
from nessai.utils import setup_logger

output = "./outdir/half_gaussian/"
logger = setup_logger(output=output)


class HalfGaussian(Model):
    """Two-dimensional Gaussian with a bound at y=0."""

    def __init__(self):
        self.names = ["x", "y"]
        self.bounds = {"x": [-10, 10], "y": [0, 10]}

    def log_prior(self, x):
        """Log-prior"""
        log_p = np.log(self.in_bounds(x), dtype="float")
        for bounds in self.bounds.values():
            log_p -= np.log(bounds[1] - bounds[0])
        return log_p

    def log_likelihood(self, x):
        """Log-likelihood"""
        log_l = np.zeros(x.size)
        for n in self.names:
            log_l += norm.logpdf(x[n])
        return log_l


# Enable boundary inversion for 'y' since we expect the posterior to rail
# against the bounds
fs = FlowSampler(
    HalfGaussian(),
    output=output,
    resume=False,
    seed=1234,
    boundary_inversion=["y"],
)

# And go!
fs.run()



# -*- coding: utf-8 -*-
import pytest
from unittest.mock import create_autospec

from nessai.proposal import FlowProposal


@pytest.fixture()
def proposal():
    proposal = create_autospec(FlowProposal)
    proposal._initialised = False
    proposal.accumulate_weights = False
    return proposal



#!/usr/bin/env python

# Example of using nessai with an unbounded prior.

import numpy as np
from scipy.stats import norm

from nessai.flowsampler import FlowSampler
from nessai.livepoint import dict_to_live_points
from nessai.model import Model
from nessai.utils import setup_logger

output = "./outdir/unbounded_prior/"
logger = setup_logger(output=output)

# We define the model as usual but also need to redefine `new_point` since by
# default this tries to draw from within the prior bounds which will fail if
# the bounds are +/- inf.


class GaussianModel(Model):
    """A simple two-dimensional Gaussian Model.

    The prior is Uniform[-10, 10] on x and a unit-Gaussian on y.
    """

    def __init__(self):
        # Names of parameters to sample
        self.names = ["x", "y"]
        # Prior bounds for each parameter
        self.bounds = {"x": [-10, 10], "y": [-np.inf, np.inf]}

    def log_prior(self, x):
        """
        Returns the log-prior.

        Checks if the points are in bounds.
        """
        # Check if the values are in the priors bounds, will return -inf if not
        log_p = np.log(self.in_bounds(x), dtype="float")
        # Uniform on x
        log_p -= np.log(self.bounds["x"][1] - self.bounds["x"][0])
        # Gaussian on y
        log_p += norm(scale=5).logpdf(x["y"])
        return log_p

    def new_point(self, N=1):
        """Draw n points.

        This is used for the initial sampling. Points do not need to be drawn
        from the exact prior but algorithm will be more efficient if they are.
        """
        # There are various ways to create live points in nessai, such as
        # from dictionaries and numpy arrays. See nessai.livepoint for options
        d = {
            "x": np.random.uniform(
                self.bounds["x"][0], self.bounds["x"][1], N
            ),
            "y": norm(scale=5).rvs(size=N),
        }
        return dict_to_live_points(d)

    def log_likelihood(self, x):
        """Returns the log-likelihood.

        In this example we use a simple Gaussian likelihood.
        """
        log_l = np.zeros(x.size)
        for n in self.names:
            log_l += norm.logpdf(x[n])
        return log_l


# The sampler can then be configure as usual
fs = FlowSampler(
    GaussianModel(),
    output=output,
    resume=False,
    seed=1234,
)

# And go!
fs.run()



"""Tests for the sorting utilities"""

from dataclasses import dataclass
from typing import List

from nessai.utils.sorting import sort_reparameterisations
import pytest


@dataclass
class Reparameterisation:
    """A dataclass that mocks the Reparameterisation class"""

    name: str
    parameters: List[str]
    requires: List[str]


def test_sorting():
    """Test a basic sorting example"""
    r0 = Reparameterisation("1", ["a"], requires=["b", "c"])
    r1 = Reparameterisation("2", ["b"], requires=["c"])
    r2 = Reparameterisation("3", ["c"], requires=[])

    out = sort_reparameterisations([r0, r1, r2])
    print([o.name for o in out])
    assert out == [r2, r1, r0]


def test_sorting_existing():
    """Test sorting when there are existing parameters"""
    r0 = Reparameterisation("1", ["a"], requires=["c"])
    r1 = Reparameterisation("2", ["b"], requires=[])
    r2 = Reparameterisation("3", ["c"], requires=["d"])

    out = sort_reparameterisations([r0, r1, r2], existing_parameters=["d"])
    print([o.name for o in out])
    assert out == [r1, r2, r0]


def test_sorting_error():
    """Assert an errors is raised if there is an invalid requirement"""
    r0 = Reparameterisation("1", ["a"], requires=[])
    r1 = Reparameterisation("2", ["b"], requires=["c"])
    with pytest.raises(ValueError, match=r".* are not known \(\['a', 'b'\]\)"):
        sort_reparameterisations([r0, r1])



# -*- coding: utf-8 -*-
"""
Test the functions related to when the flow should be trained or reset and
training itself.
"""
import datetime
import numpy as np
import pytest
from unittest.mock import MagicMock
from nessai.samplers.nestedsampler import NestedSampler


def test_configure_flow_reset_false(sampler):
    """Assert the attributes evaluate to false if the inputs are false"""
    NestedSampler.configure_flow_reset(sampler, False, False, False)
    assert not sampler.reset_weights
    assert not sampler.reset_permutations
    assert not sampler.reset_flow


@pytest.mark.parametrize("weights", [10, 5.0])
@pytest.mark.parametrize("permutations", [10, 5.0])
def test_configure_flow_reset(sampler, weights, permutations):
    """Assert the attributes evaluate to false if the inputs are false"""
    NestedSampler.configure_flow_reset(sampler, weights, permutations, False)
    assert sampler.reset_weights == float(weights)
    assert sampler.reset_permutations == float(permutations)


@pytest.mark.parametrize("flow", [10, 5.0])
def test_configure_flow_reset_flow(sampler, flow):
    """Assert reset_flow overwrites the other values"""
    NestedSampler.configure_flow_reset(sampler, 2, 4, flow)
    assert sampler.reset_flow == float(flow)
    assert sampler.reset_weights == float(flow)
    assert sampler.reset_permutations == float(flow)


def test_configure_flow_reset_error_weights(sampler):
    """Assert an error is raised in the weights input is invalid"""
    with pytest.raises(TypeError) as excinfo:
        NestedSampler.configure_flow_reset(sampler, None, 5, 5)
    assert "weights` must be" in str(excinfo.value)


def test_configure_flow_reset_error_permutations(sampler):
    """Assert an error is raised in the permutations input is invalid"""
    with pytest.raises(TypeError) as excinfo:
        NestedSampler.configure_flow_reset(sampler, 5, None, 5)
    assert "permutations` must be" in str(excinfo.value)


def test_configure_flow_reset_error_flow(sampler):
    """Assert an error is raised if the flow input is invalid"""
    with pytest.raises(TypeError) as excinfo:
        NestedSampler.configure_flow_reset(sampler, 5, 5, None)
    assert "`reset_flow` must be" in str(excinfo.value)


def test_check_training_not_completed_training(sampler):
    """
    Assert the flow is forced to train if training did not complete when
    the sampler was checkpointed.
    """
    sampler.completed_training = False
    train, force = NestedSampler.check_training(sampler)
    assert train is True
    assert force is True


def test_check_training_train_on_empty(sampler):
    """
    Assert the flow is forced to train if training the pool is empty and
    `train_on_empty` is true but the proposal was not in the process of
    popluating.
    """
    sampler.completed_training = True
    sampler.train_on_empty = True
    sampler.proposal = MagicMock()
    sampler.proposal.populated = False
    sampler.proposal.populating = False
    train, force = NestedSampler.check_training(sampler)
    assert train is True
    assert force is True


def test_check_training_acceptance(sampler):
    """
    Assert that training will be true but not forced if the acceptance
    threshold is met and retraining on acceptance is enabled.
    """
    sampler.completed_training = True
    sampler.train_on_empty = True
    sampler.proposal = MagicMock()
    sampler.proposal.populated = True
    sampler.proposal.populating = False
    sampler.acceptance_threshold = 0.1
    sampler.mean_block_acceptance = 0.01
    sampler.retrain_acceptance = True
    train, force = NestedSampler.check_training(sampler)
    assert train is True
    assert force is False


def test_check_training_iteration(sampler):
    """
    Assert that training will be true but not forced if a training iteration
    is reached (n iterations have passed since last updated).
    """
    sampler.completed_training = True
    sampler.train_on_empty = True
    sampler.proposal = MagicMock()
    sampler.proposal.populated = True
    sampler.proposal.populating = False
    sampler.acceptance_threshold = 0.1
    sampler.mean_block_acceptance = 0.2
    sampler.retrain_acceptance = False
    sampler.iteration = 3521
    sampler.last_updated = 2521
    sampler.training_frequency = 1000
    train, force = NestedSampler.check_training(sampler)
    assert train is True
    assert force is False


@pytest.mark.parametrize(
    "config",
    [
        dict(),
        dict(train_on_empty=False, populated=False),
        dict(train_on_empty=True, populated=False, populating=True),
        dict(
            mean_acceptance=0.01,
            acceptance_threshold=0.5,
            retrain_acceptance=False,
        ),
        dict(
            mean_acceptance=0.5,
            acceptance_threshold=0.01,
            retrain_acceptance=True,
        ),
        dict(iteration=800, last_updated=0, training_frequency=801),
    ],
)
def test_check_training_false(sampler, config):
    """
    Test a range of different scenarios that should all not start training.
    """
    sampler.completed_training = True
    sampler.train_on_empty = config.get("train_on_empty", False)
    sampler.proposal = MagicMock()
    sampler.proposal.populated = config.get("populated", False)
    sampler.proposal.populating = config.get("populating", False)
    sampler.acceptance_threshold = config.get("acceptance_threshold", 0.1)
    sampler.mean_block_acceptance = config.get("mean_acceptance", 0.2)
    sampler.retrain_acceptance = config.get("retrain_acceptance", False)
    sampler.iteration = config.get("iteration", 3000)
    sampler.last_updated = config.get("last_updated", 2500)
    sampler.training_frequency = config.get("training_frequency", 1000)
    train, force = NestedSampler.check_training(sampler)
    assert train is False
    assert force is False


@pytest.mark.parametrize("training_count", [10, 100])
def test_check_flow_model_reset_weights(sampler, training_count):
    """Assert flow model only weights are reset"""
    sampler.proposal = MagicMock()
    sampler.proposal.reset_model_weights = MagicMock()
    sampler.reset_acceptance = False
    sampler.reset_weights = 10
    sampler.reset_permutations = 0
    sampler.proposal.training_count = training_count

    NestedSampler.check_flow_model_reset(sampler)

    sampler.proposal.reset_model_weights.assert_called_once_with(
        weights=True,
        permutations=False,
    )


@pytest.mark.parametrize("training_count", [10, 100])
def test_check_flow_model_reset_permutations(sampler, training_count):
    """Assert flow model only permutations are reset"""
    sampler.proposal = MagicMock()
    sampler.proposal.reset_model_weights = MagicMock()
    sampler.reset_acceptance = False
    sampler.reset_weights = 0
    sampler.reset_permutations = 10
    sampler.proposal.training_count = training_count

    NestedSampler.check_flow_model_reset(sampler)

    sampler.proposal.reset_model_weights.assert_called_once_with(
        weights=False, permutations=True
    )


@pytest.mark.parametrize("training_count", [10, 100])
def test_check_flow_model_reset_both(sampler, training_count):
    """Assert flow model only permutations are reset"""
    sampler.proposal = MagicMock()
    sampler.proposal.reset_model_weights = MagicMock()
    sampler.reset_acceptance = False
    sampler.reset_weights = 10
    sampler.reset_permutations = 10
    sampler.proposal.training_count = training_count

    NestedSampler.check_flow_model_reset(sampler)

    sampler.proposal.reset_model_weights.assert_called_once_with(
        weights=True,
        permutations=True,
    )


def test_check_flow_model_reset_acceptance(sampler):
    """
    Assert flow model is reset based on acceptance is reset_acceptance is True.
    """
    sampler.proposal = MagicMock()
    sampler.proposal.reset_model_weights = MagicMock()
    sampler.reset_acceptance = True
    sampler.mean_block_acceptance = 0.1
    sampler.acceptance_threshold = 0.5
    sampler.proposal.training_count = 1

    NestedSampler.check_flow_model_reset(sampler)

    sampler.proposal.reset_model_weights.assert_called_once_with(
        weights=True, permutations=True
    )


def test_check_flow_model_reset_not_trained(sampler):
    """
    Verify that the flow model is not reset if it has never been trained.
    """
    sampler.proposal = MagicMock()
    sampler.proposal.reset_model_weights = MagicMock()
    sampler.proposal.training_count = 0

    NestedSampler.check_flow_model_reset(sampler)

    sampler.proposal.reset_model_weights.assert_not_called()


def test_train_proposal_not_training(sampler):
    """Verify the proposal is not trained it has not 'cooled down'"""
    sampler.proposal = MagicMock()
    sampler.proposal.train = MagicMock()
    sampler.iteration = 100
    sampler.last_updated = 90
    sampler.cooldown = 20
    NestedSampler.train_proposal(sampler, force=False)
    sampler.proposal.train.assert_not_called()


def test_train_proposal(sampler, wait):
    """Verify the proposal is trained"""
    sampler.proposal = MagicMock()
    sampler.proposal.train = MagicMock(side_effect=wait)
    sampler.check_flow_model_reset = MagicMock()
    sampler.checkpoint = MagicMock()
    sampler.iteration = 100
    sampler.last_updated = 90
    sampler.cooldown = 20
    sampler.memory = False
    sampler.training_time = datetime.timedelta()
    sampler.history = dict(training_iterations=[])
    sampler.live_points = np.arange(10)
    sampler.checkpoint_on_training = True
    sampler.block_iteration = 10
    sampler.block_acceptance = 0.5

    NestedSampler.train_proposal(sampler, force=True)

    sampler.check_flow_model_reset.assert_called_once()
    sampler.proposal.train.assert_called_once()
    sampler.checkpoint.assert_called_once_with(periodic=True)

    assert sampler.history["training_iterations"] == [100]
    assert sampler.training_time.total_seconds() > 0
    assert sampler.completed_training is True
    assert sampler.block_iteration == 0
    assert sampler.block_acceptance == 0


def test_train_proposal_memory(sampler, wait):
    """Verify the proposal is trained with memory"""
    sampler.proposal = MagicMock()
    sampler.proposal.train = MagicMock(side_effect=wait)
    sampler.check_flow_model_reset = MagicMock()
    sampler.checkpoint = MagicMock()
    sampler.iteration = 100
    sampler.last_updated = 90
    sampler.cooldown = 20
    sampler.memory = 2
    sampler.training_time = datetime.timedelta()
    sampler.history = dict(training_iterations=[])
    sampler.nested_samples = np.arange(5)
    sampler.live_points = np.arange(5, 10)
    sampler.checkpoint_on_training = True
    sampler.block_iteration = 10
    sampler.block_acceptance = 0.5

    NestedSampler.train_proposal(sampler, force=True)

    sampler.check_flow_model_reset.assert_called_once()
    sampler.checkpoint.assert_called_once_with(periodic=True)
    sampler.proposal.train.assert_called_once()

    np.testing.assert_array_equal(
        sampler.proposal.train.call_args[0], np.array([[5, 6, 7, 8, 9, 3, 4]])
    )

    assert sampler.history["training_iterations"] == [100]
    assert sampler.training_time.total_seconds() > 0
    assert sampler.completed_training is True
    assert sampler.block_iteration == 0
    assert sampler.block_acceptance == 0



# -*- coding: utf-8 -*-
"""Test methods related to popluation of the proposal after training"""
import datetime
from functools import partial
import os

import numpy as np
import pytest
from unittest.mock import MagicMock, Mock, patch, call

from nessai import config
from nessai.proposal import FlowProposal
from nessai.livepoint import get_dtype, numpy_array_to_live_points
from nessai.utils.testing import assert_structured_arrays_equal


@pytest.fixture()
def z():
    return np.random.randn(2, 2)


@pytest.fixture()
def x(z):
    return numpy_array_to_live_points(np.random.randn(*z.shape), ["x", "y"])


@pytest.fixture()
def log_q(x):
    return np.random.randn(x.size)


def test_log_prior_wo_reparameterisation(proposal, x):
    """Test the lop prior method"""
    log_prior = -np.ones(x.size)
    proposal._reparameterisation = None
    proposal.model = MagicMock()
    proposal.model.batch_evaluate_log_prior = MagicMock(return_value=log_prior)

    log_prior_out = FlowProposal.log_prior(proposal, x)

    assert np.array_equal(log_prior, log_prior_out)
    proposal.model.batch_evaluate_log_prior.assert_called_once_with(x)


def test_log_prior_w_reparameterisation(proposal, x):
    """Test the lop prior method with reparameterisations"""
    log_prior = -np.ones(x.size)
    proposal._reparameterisation = MagicMock()
    proposal._reparameterisation.log_prior = MagicMock(return_value=log_prior)
    proposal.model = MagicMock()
    proposal.model.batch_evaluate_log_prior = MagicMock(
        return_value=log_prior.copy()
    )

    log_prior_out = FlowProposal.log_prior(proposal, x)

    assert np.array_equal(log_prior_out, -2 * np.ones(x.size))
    proposal._reparameterisation.log_prior.assert_called_once_with(x)
    proposal.model.batch_evaluate_log_prior.assert_called_once_with(x)


def test_prime_log_prior(proposal):
    """Make sure the prime prior raises an error by default."""
    with pytest.raises(RuntimeError) as excinfo:
        FlowProposal.x_prime_log_prior(proposal, 1.0)
    assert "Prime prior is not implemented" in str(excinfo.value)


@pytest.mark.parametrize(
    "acceptance, scale", [(0.0, 10.0), (0.5, 2.0), (0.01, 10.0), (2.0, 1.0)]
)
def test_update_poolsize_scale(proposal, acceptance, scale):
    """
    Test the check the poolsize is correct adjusted based on the acceptance.
    """
    proposal.max_poolsize_scale = 10.0
    FlowProposal.update_poolsize_scale(proposal, acceptance)
    assert proposal._poolsize_scale == scale


def test_compute_weights(proposal, x, log_q):
    """Test method for computing rejection sampling weights"""
    proposal.use_x_prime_prior = False
    proposal.log_prior = MagicMock(return_value=-np.ones(x.size))
    log_w = FlowProposal.compute_weights(proposal, x, log_q)

    proposal.log_prior.assert_called_once_with(x)
    out = -1 - log_q
    np.testing.assert_array_equal(log_w, out)


def test_compute_weights_return_prior(proposal, x, log_q):
    """Assert prior is returned"""
    proposal.use_x_prime_prior = False
    log_p = -np.ones(x.size)
    proposal.log_prior = MagicMock(return_value=log_p)
    log_w, log_p_out = FlowProposal.compute_weights(
        proposal, x, log_q, return_log_prior=True
    )

    proposal.log_prior.assert_called_once_with(x)
    expected = -1 - log_q
    np.testing.assert_array_equal(log_w, expected)
    assert log_p_out is log_p


def test_compute_weights_prime_prior(proposal, x, log_q):
    """Test method for computing rejection sampling weights with the prime
    prior.
    """
    proposal.use_x_prime_prior = True
    proposal.x_prime_log_prior = MagicMock(return_value=-np.ones(x.size))
    log_w = FlowProposal.compute_weights(proposal, x, log_q)

    proposal.x_prime_log_prior.assert_called_once_with(x)
    out = -1 - log_q
    np.testing.assert_array_equal(log_w, out)


@patch("numpy.random.rand", return_value=np.array([0.1, 0.9]))
def test_rejection_sampling(proposal, z, x, log_q):
    """Test rejection sampling method."""
    proposal.use_x_prime_prior = False
    proposal.truncate = False
    proposal.backward_pass = MagicMock(return_value=(x, log_q, z))
    log_w = np.log(np.array([0.5, 0.5]))
    proposal.compute_weights = MagicMock(return_value=log_w)

    z_out, x_out = FlowProposal.rejection_sampling(proposal, z)

    proposal.backward_pass.assert_called_once_with(
        z, rescale=True, return_z=True, discard_nans=False
    )
    proposal.compute_weights.assert_called_once()
    assert x_out.size == 1
    assert z_out.shape == (1, 2)
    assert_structured_arrays_equal(x_out[0], x[0])
    assert np.array_equal(z_out[0], z[0])


def test_rejection_sampling_empty(proposal, z):
    """Test rejection sampling method if no valid points are produced by
    `backwards_pass`
    """
    proposal.use_x_prime_prior = False
    proposal.truncate = False
    proposal.backward_pass = MagicMock(
        return_value=(np.array([]), np.array([]), np.array([]))
    )

    z_out, x_out = FlowProposal.rejection_sampling(proposal, z)

    assert x_out.size == 0
    assert z_out.size == 0


@patch("numpy.random.rand", return_value=np.array([0.1]))
def test_rejection_sampling_truncate(proposal, z, x):
    """Test rejection sampling method with truncation"""
    proposal.use_x_prime_prior = False
    proposal.truncate = True
    log_q = np.array([0.0, 1.0])
    proposal.backward_pass = MagicMock(return_value=(x, log_q, z))
    min_log_q = 0.5
    log_w = np.log(np.array([0.5]))
    proposal.compute_weights = MagicMock(return_value=log_w)

    z_out, x_out = FlowProposal.rejection_sampling(
        proposal,
        z,
        min_log_q=min_log_q,
    )

    proposal.backward_pass.assert_called_once_with(
        z, rescale=True, return_z=True, discard_nans=False
    )
    proposal.compute_weights.assert_called_once()
    assert x_out.size == 1
    assert z_out.shape == (1, 2)
    assert_structured_arrays_equal(x_out[0], x[1])
    assert np.array_equal(z_out[0], z[1])


def test_compute_acceptance(proposal):
    """Test the compute_acceptance method"""
    proposal.samples = np.arange(1, 11, dtype=float).view([("logL", "f8")])
    acc = FlowProposal.compute_acceptance(proposal, 5.0)
    assert acc == 0.5


def test_convert_to_samples(proposal):
    """Test convert to sample without the prime prior"""
    samples = numpy_array_to_live_points(np.random.randn(10, 2), ["x", "y"])
    proposal.use_x_prime_prior = False
    proposal.model = MagicMock()
    proposal.model.names = ["x"]
    proposal.model.batch_evaluate_log_prior = MagicMock(
        return_value=np.ones(10)
    )

    out_samples = FlowProposal.convert_to_samples(proposal, samples, plot=True)

    assert out_samples.dtype.names == ("x",) + tuple(
        config.livepoints.non_sampling_parameters
    )


@patch("nessai.proposal.flowproposal.plot_1d_comparison")
def test_convert_to_samples_with_prime(mock_plot, proposal):
    """Test convert to sample with the prime prior"""
    samples = numpy_array_to_live_points(np.random.randn(10, 2), ["x", "y"])
    proposal.use_x_prime_prior = True
    proposal.model = MagicMock()
    proposal.model.names = ["x"]
    proposal.model.batch_evaluate_log_prior = MagicMock(
        return_value=np.ones(10)
    )
    proposal._plot_pool = True
    proposal.training_data_prime = "data"
    proposal.output = os.getcwd()
    proposal.populated_count = 1
    proposal.inverse_rescale = MagicMock(return_value=(samples, None))

    out_samples = FlowProposal.convert_to_samples(proposal, samples, plot=True)

    mock_plot.assert_called_once_with(
        "data",
        samples,
        labels=["live points", "pool"],
        filename=os.path.join(os.getcwd(), "pool_prime_1.png"),
    )
    proposal.inverse_rescale.assert_called_once()
    assert out_samples.dtype.names == ("x",) + tuple(
        config.livepoints.non_sampling_parameters
    )


def test_get_alt_distribution_truncated_gaussian(proposal):
    """
    Test getting the alternative distribution for the default latent prior, the
    truncated Gaussian with var=1. This should return None.
    """
    proposal.draw_latent_kwargs = {}
    proposal.latent_prior = "truncated_gaussian"
    dist = FlowProposal.get_alt_distribution(proposal)
    assert dist is None


@pytest.mark.parametrize("prior", ["uniform_nsphere", "uniform_nball"])
def test_get_alt_distribution_uniform(proposal, prior):
    """
    Test getting the alternative distribution for priors that are uniform in
    the n-ball.
    """
    proposal.latent_prior = prior
    proposal.dims = 2
    proposal.r = 2.0
    proposal.fuzz = 1.2
    proposal.flow = Mock()
    proposal.flow.device = "cpu"
    with patch(
        "nessai.proposal.flowproposal.get_uniform_distribution"
    ) as mock:
        dist = FlowProposal.get_alt_distribution(proposal)

    assert dist is not None
    mock.assert_called_once_with(2, 2.4, device="cpu")


def test_radius(proposal):
    """Test computing the radius"""
    z = np.array([[1, 2, 3], [0, 1, 2]])
    expected_r = np.sqrt(14)
    r = FlowProposal.radius(proposal, z)
    np.testing.assert_equal(r, expected_r)


def test_radius_w_log_q(proposal):
    """Test computing the radius with log_q"""
    z = np.array([[1, 2, 3], [0, 1, 2]])
    log_q = np.array([1, 2])
    expected_r = np.sqrt(14)
    r, log_q_r = FlowProposal.radius(proposal, z, log_q)
    assert r == expected_r
    assert log_q_r == log_q[0]


def test_check_prior_bounds(proposal):
    """Test the check prior bounds method."""
    x = numpy_array_to_live_points(np.arange(10)[:, np.newaxis], ["x"])
    y = np.arange(10)
    proposal.model = Mock()
    proposal.model.in_bounds = MagicMock(
        return_value=np.array(6 * [True] + 4 * [False])
    )
    x_out, y_out = FlowProposal.check_prior_bounds(proposal, x, y)

    assert_structured_arrays_equal(x_out, x[:6])
    np.testing.assert_array_equal(y_out, y[:6])


def test_prep_latent_prior_truncated(proposal):
    """Assert prep latent prior calls the correct values"""

    proposal.latent_prior = "truncated_gaussian"
    proposal.dims = 2
    proposal.r = 3.0
    proposal.fuzz = 1.2
    dist = MagicMock()
    dist.sample = MagicMock()

    with patch(
        "nessai.proposal.flowproposal.NDimensionalTruncatedGaussian",
        return_value=dist,
    ) as mock_dist:
        FlowProposal.prep_latent_prior(proposal)

    mock_dist.assert_called_once_with(2, 3.0, fuzz=1.2)

    assert proposal._populate_dist is dist
    assert proposal._draw_func is dist.sample


def test_prep_latent_prior_other(proposal):
    """Assert partial acts as expected"""
    proposal.latent_prior = "gaussian"
    proposal.dims = 2
    proposal.r = 3.0
    proposal.fuzz = 1.2

    def draw(dims, N=None, r=None, fuzz=None):
        return np.zeros((N, dims))

    proposal._draw_latent_prior = draw

    with patch(
        "nessai.proposal.flowproposal.partial", side_effect=partial
    ) as mock_partial:
        FlowProposal.prep_latent_prior(proposal)

    mock_partial.assert_called_once_with(draw, dims=2, r=3.0, fuzz=1.2)

    assert proposal._draw_func(N=10).shape == (10, 2)


def test_prep_latent_prior_flow(proposal):
    proposal.latent_prior = "flow"
    proposal.flow = MagicMock()
    proposal.flow.sample_latent_distribution = MagicMock()
    FlowProposal.prep_latent_prior(proposal)
    proposal._draw_func(10)
    proposal.flow.sample_latent_distribution.assert_called_once_with(10)


def test_draw_latent_prior(proposal):
    proposal._draw_func = MagicMock(return_value=[1, 2])
    out = FlowProposal.draw_latent_prior(proposal, 2)
    proposal._draw_func.assert_called_once_with(N=2)
    assert out == [1, 2]


@pytest.mark.parametrize("check_acceptance", [False, True])
@pytest.mark.parametrize("indices", [[], [1]])
@pytest.mark.parametrize("r", [None, 1.0])
@pytest.mark.parametrize(
    "max_radius, min_radius",
    [(0.5, None), (None, 1.5), (None, None)],
)
def test_populate_accumulate_weights(
    proposal, check_acceptance, indices, r, min_radius, max_radius, wait
):
    """Test the main populate method"""
    n_dims = 2
    poolsize = 10
    drawsize = 5
    names = ["x", "y"]
    worst_point = np.array(
        [[1, 2, 3]], dtype=[("x", "f8"), ("y", "f8"), ("logL", "f8")]
    )
    worst_z = np.random.randn(1, n_dims)
    z = [
        np.random.randn(drawsize, n_dims),
        np.random.randn(drawsize, n_dims),
        np.random.randn(drawsize, n_dims),
    ]
    x = [
        numpy_array_to_live_points(np.random.randn(drawsize, n_dims), names),
        numpy_array_to_live_points(np.random.randn(drawsize, n_dims), names),
        numpy_array_to_live_points(np.random.randn(drawsize, n_dims), names),
    ]
    log_q = [
        np.log(np.random.rand(drawsize)),
        np.log(np.random.rand(drawsize)),
        np.log(np.random.rand(drawsize)),
    ]
    log_w = [
        np.log(np.concatenate([np.ones(drawsize - 1), np.zeros(1)])),
        np.log(np.concatenate([np.ones(drawsize - 1), np.zeros(1)])),
        np.log(np.concatenate([np.ones(drawsize - 1), np.zeros(1)])),
    ]
    # Control rejection sampling using log_w
    rand_u = 0.5 * np.ones(3 * drawsize)

    log_l = np.random.rand(poolsize)
    log_p = np.random.rand(poolsize)

    r_flow = 1.0

    if r is None:
        r_out = r_flow
        if min_radius is not None:
            r_out = max(r_out, min_radius)
        if max_radius is not None:
            r_out = min(r_out, max_radius)
    else:
        r_out = r

    proposal.population_time = datetime.timedelta()
    proposal.initialised = True
    proposal.max_radius = max_radius
    proposal.dims = n_dims
    proposal.poolsize = poolsize
    proposal.drawsize = drawsize
    proposal.min_radius = min_radius
    proposal.fuzz = 1.0
    proposal.indices = indices
    proposal.acceptance = [0.7]
    proposal.keep_samples = False
    proposal.fixed_radius = False
    proposal.compute_radius_with_all = False
    proposal.check_acceptance = check_acceptance
    proposal._plot_pool = True
    proposal.populated_count = 1
    proposal.population_dtype = get_dtype(names)
    proposal.truncate_log_q = False
    proposal.use_x_prime_prior = False
    proposal.accumulate_weights = True

    proposal.forward_pass = MagicMock(return_value=(worst_z, np.nan))
    proposal.backward_pass = MagicMock(side_effect=zip(x, log_q))
    proposal.radius = MagicMock(return_value=r_flow)
    proposal.get_alt_distribution = MagicMock(return_value=None)
    proposal.prep_latent_prior = MagicMock()
    proposal.draw_latent_prior = MagicMock(side_effect=z)
    proposal.compute_weights = MagicMock(side_effect=log_w)
    proposal.compute_acceptance = MagicMock(return_value=0.8)
    proposal.model = MagicMock()
    proposal.model.batch_evaluate_log_likelihood = MagicMock(
        return_value=log_l
    )

    def convert_to_samples(samples, plot):
        samples["logP"] = log_p
        # wait for windows
        wait()
        return samples

    proposal.plot_pool = MagicMock()
    proposal.convert_to_samples = MagicMock(side_effect=convert_to_samples)

    x_empty = np.empty(0, dtype=proposal.population_dtype)
    with patch(
        "nessai.proposal.flowproposal.empty_structured_array",
        return_value=x_empty,
    ) as mock_empty, patch(
        "numpy.random.rand", return_value=rand_u
    ) as mock_rand:
        FlowProposal.populate(
            proposal, worst_point, N=poolsize, plot=True, r=r
        )

    mock_empty.assert_called_once_with(
        0,
        dtype=proposal.population_dtype,
    )
    mock_rand.assert_called_once_with(3 * drawsize)

    if r is None:
        proposal.forward_pass.assert_called_once_with(
            worst_point,
            rescale=True,
            compute_radius=True,
        )
        proposal.radius.assert_called_once_with(worst_z)
    else:
        assert proposal.r is r

    assert proposal.r == r_out

    proposal.prep_latent_prior.assert_called_once()

    draw_calls = 3 * [call(5)]
    proposal.draw_latent_prior.assert_has_calls(draw_calls)

    backwards_calls = [call(zz, rescale=True) for zz in z]
    proposal.backward_pass.assert_has_calls(backwards_calls)

    compute_weights_calls = [call(xx, lq) for xx, lq in zip(x, log_q)]
    proposal.compute_weights.assert_has_calls(compute_weights_calls)

    proposal.plot_pool.assert_called_once()
    proposal.convert_to_samples.assert_called_once()
    assert_structured_arrays_equal(
        proposal.convert_to_samples.call_args[0][0], proposal.x
    )
    assert proposal.convert_to_samples.call_args[1]["plot"] is True

    assert proposal.population_acceptance == (12 / 15)
    assert proposal.populated_count == 2
    assert proposal.populated is True
    assert proposal.x.size == 10

    if check_acceptance:
        proposal.compute_acceptance.assert_called()
        assert proposal.acceptance == [0.7, 0.8]
    else:
        proposal.compute_acceptance.assert_not_called()

    proposal.model.batch_evaluate_log_likelihood.assert_called_once_with(
        proposal.samples
    )
    np.testing.assert_array_equal(proposal.samples["logL"], log_l)

    assert proposal.population_time.total_seconds() > 0.0


@pytest.mark.parametrize("check_acceptance", [False, True])
@pytest.mark.parametrize("indices", [[], [1]])
@pytest.mark.parametrize("r", [None, 1.0])
@pytest.mark.parametrize(
    "max_radius, min_radius",
    [(0.5, None), (None, 1.5), (None, None)],
)
def test_populate_not_accumulate_weights(
    proposal, check_acceptance, indices, r, min_radius, max_radius, wait
):
    """Test the main populate method"""
    n_dims = 2
    poolsize = 10
    drawsize = 5
    names = ["x", "y"]
    worst_point = np.array(
        [[1, 2, 3]], dtype=[("x", "f8"), ("y", "f8"), ("logL", "f8")]
    )
    worst_z = np.random.randn(1, n_dims)
    z = [
        np.random.randn(drawsize, n_dims),
        np.random.randn(drawsize, n_dims),
        np.random.randn(drawsize, n_dims),
    ]
    x = [
        numpy_array_to_live_points(np.random.randn(drawsize, n_dims), names),
        numpy_array_to_live_points(np.random.randn(drawsize, n_dims), names),
        numpy_array_to_live_points(np.random.randn(drawsize, n_dims), names),
    ]
    log_q = [
        np.log(np.random.rand(drawsize)),
        np.log(np.random.rand(drawsize)),
        np.log(np.random.rand(drawsize)),
    ]
    log_w = [
        np.log(np.concatenate([np.ones(drawsize - 1), np.zeros(1)])),
        np.log(np.concatenate([np.ones(drawsize - 1), np.zeros(1)])),
        np.log(np.concatenate([np.ones(drawsize - 1), np.zeros(1)])),
    ]
    # Control rejection sampling using log_w
    rand_u = [
        0.5 * np.ones(drawsize),
        0.5 * np.ones(drawsize),
        0.5 * np.ones(drawsize),
    ]

    log_l = np.random.rand(poolsize)
    log_p = np.random.rand(poolsize)

    r_flow = 1.0

    if r is None:
        r_out = r_flow
        if min_radius is not None:
            r_out = max(r_out, min_radius)
        if max_radius is not None:
            r_out = min(r_out, max_radius)
    else:
        r_out = r

    proposal.population_time = datetime.timedelta()
    proposal.initialised = True
    proposal.max_radius = max_radius
    proposal.dims = n_dims
    proposal.poolsize = poolsize
    proposal.drawsize = drawsize
    proposal.min_radius = min_radius
    proposal.fuzz = 1.0
    proposal.indices = indices
    proposal.acceptance = [0.7]
    proposal.keep_samples = False
    proposal.fixed_radius = False
    proposal.compute_radius_with_all = False
    proposal.check_acceptance = check_acceptance
    proposal._plot_pool = True
    proposal.populated_count = 1
    proposal.population_dtype = get_dtype(names)
    proposal.truncate_log_q = False
    proposal.use_x_prime_prior = False
    proposal.accumulate_weights = False

    proposal.forward_pass = MagicMock(return_value=(worst_z, np.nan))
    proposal.backward_pass = MagicMock(side_effect=zip(x, log_q))
    proposal.radius = MagicMock(return_value=r_flow)
    proposal.get_alt_distribution = MagicMock(return_value=None)
    proposal.prep_latent_prior = MagicMock()
    proposal.draw_latent_prior = MagicMock(side_effect=z)
    proposal.compute_weights = MagicMock(side_effect=log_w)
    proposal.compute_acceptance = MagicMock(return_value=0.8)
    proposal.model = MagicMock()
    proposal.model.batch_evaluate_log_likelihood = MagicMock(
        return_value=log_l
    )

    def convert_to_samples(samples, plot):
        samples["logP"] = log_p
        # wait for windows
        wait()
        return samples

    proposal.plot_pool = MagicMock()
    proposal.convert_to_samples = MagicMock(side_effect=convert_to_samples)

    x_empty = np.empty(poolsize, dtype=proposal.population_dtype)
    with patch(
        "nessai.proposal.flowproposal.empty_structured_array",
        return_value=x_empty,
    ) as mock_empty, patch(
        "numpy.random.rand", side_effect=rand_u
    ) as mock_rand:
        FlowProposal.populate(
            proposal, worst_point, N=poolsize, plot=True, r=r
        )

    mock_empty.assert_called_once_with(
        poolsize,
        dtype=proposal.population_dtype,
    )
    mock_rand.assert_has_calls(3 * [call(drawsize)])

    if r is None:
        proposal.forward_pass.assert_called_once_with(
            worst_point,
            rescale=True,
            compute_radius=True,
        )
        proposal.radius.assert_called_once_with(worst_z)
    else:
        assert proposal.r is r

    assert proposal.r == r_out

    proposal.prep_latent_prior.assert_called_once()

    draw_calls = 3 * [call(5)]
    proposal.draw_latent_prior.assert_has_calls(draw_calls)

    backwards_calls = [call(zz, rescale=True) for zz in z]
    proposal.backward_pass.assert_has_calls(backwards_calls)

    compute_weights_calls = [call(xx, lq) for xx, lq in zip(x, log_q)]
    proposal.compute_weights.assert_has_calls(compute_weights_calls)

    proposal.plot_pool.assert_called_once()
    proposal.convert_to_samples.assert_called_once()
    assert_structured_arrays_equal(
        proposal.convert_to_samples.call_args[0][0], proposal.x
    )
    assert proposal.convert_to_samples.call_args[1]["plot"] is True

    assert proposal.population_acceptance == (12 / 15)
    assert proposal.populated_count == 2
    assert proposal.populated is True
    assert proposal.x.size == 10

    if check_acceptance:
        proposal.compute_acceptance.assert_called()
        assert proposal.acceptance == [0.7, 0.8]
    else:
        proposal.compute_acceptance.assert_not_called()

    proposal.model.batch_evaluate_log_likelihood.assert_called_once_with(
        proposal.samples
    )
    np.testing.assert_array_equal(proposal.samples["logL"], log_l)

    assert proposal.population_time.total_seconds() > 0.0


def test_populate_not_initialised(proposal):
    """Assert populate fails if the proposal is not initialised"""
    proposal.initialised = False
    with pytest.raises(RuntimeError) as excinfo:
        FlowProposal.populate(proposal, 1.0)
    assert "Proposal has not been initialised. " in str(excinfo.value)


def test_populate_truncate_log_q(proposal):
    n_dims = 2
    nlive = 8
    poolsize = 8
    drawsize = 5
    names = ["x", "y"]
    r_flow = 2.0
    worst_point = np.array(
        [[1, 2, 3]], dtype=[("x", "f8"), ("y", "f8"), ("logL", "f8")]
    )
    z = [
        np.random.randn(drawsize, n_dims),
        np.random.randn(drawsize, n_dims),
        np.random.randn(drawsize, n_dims),
    ]
    x = [
        numpy_array_to_live_points(np.random.randn(drawsize, n_dims), names),
        numpy_array_to_live_points(np.random.randn(drawsize, n_dims), names),
        numpy_array_to_live_points(np.random.randn(drawsize, n_dims), names),
    ]
    log_q = [
        np.zeros(drawsize),
        np.zeros(drawsize),
        np.zeros(drawsize),
    ]
    # This sample will be discarded because of the logq min check
    for i in range(3):
        log_q[i][-1] = np.nan_to_num(-np.inf)
    log_w = [
        np.log(np.concatenate([np.ones(drawsize - 2), np.zeros(1)])),
        np.log(np.concatenate([np.ones(drawsize - 2), np.zeros(1)])),
        np.log(np.concatenate([np.ones(drawsize - 2), np.zeros(1)])),
    ]
    # Control rejection sampling using log_w
    rand_u = 0.5 * np.ones(3 * (drawsize - 1))

    log_l = np.random.rand(poolsize)

    proposal.population_time = datetime.timedelta()
    proposal.initialised = True
    proposal.dims = n_dims
    proposal.poolsize = poolsize
    proposal.drawsize = drawsize
    proposal.fuzz = 1.0
    proposal.indices = None
    proposal.acceptance = [0.7]
    proposal.keep_samples = False
    proposal.fixed_radius = 2.0
    proposal.compute_radius_with_all = False
    proposal.check_acceptance = False
    proposal._plot_pool = False
    proposal.use_x_prime_prior = False
    proposal.populated_count = 1
    proposal.population_dtype = get_dtype(names)
    proposal.truncate_log_q = True
    proposal.training_data = numpy_array_to_live_points(
        np.random.randn(nlive, n_dims),
        names=names,
    )
    proposal.accumulate_weights = True

    log_q_live = np.zeros(nlive)
    log_q_live[-1] = -1.0

    proposal.forward_pass = MagicMock(
        return_value=(nlive * [None], log_q_live)
    )
    proposal.backward_pass = MagicMock(side_effect=zip(x, log_q))
    proposal.compute_weights = MagicMock(side_effect=log_w)
    proposal.radius = MagicMock(return_value=r_flow)
    proposal.get_alt_distribution = MagicMock(return_value=None)
    proposal.prep_latent_prior = MagicMock()
    proposal.draw_latent_prior = MagicMock(side_effect=z)
    proposal.rejection_sampling = MagicMock(
        side_effect=[(a[:-1], b[:-1]) for a, b in zip(z, x)]
    )
    proposal.compute_acceptance = MagicMock(return_value=0.8)
    proposal.model = MagicMock()
    proposal.model.batch_evaluate_log_likelihood = MagicMock(
        return_value=log_l
    )

    proposal.convert_to_samples = MagicMock(
        side_effect=lambda *args, **kwargs: args[0]
    )

    x_empty = np.empty(0, dtype=proposal.population_dtype)
    with patch(
        "nessai.proposal.flowproposal.empty_structured_array",
        return_value=x_empty,
    ) as mock_empty, patch(
        "numpy.random.rand", return_value=rand_u
    ) as mock_rand:
        FlowProposal.populate(proposal, worst_point, N=poolsize, plot=False)

    mock_empty.assert_called_once_with(
        0,
        dtype=proposal.population_dtype,
    )
    mock_rand.assert_called_once_with(3 * drawsize - 3)

    assert proposal.population_acceptance == (9 / 15)

    proposal.forward_pass.assert_called_once_with(proposal.training_data)

    backwards_calls = [call(zz, rescale=True) for zz in z]
    proposal.backward_pass.assert_has_calls(backwards_calls)

    compute_weights_calls = [(xx[:-1], lq[:-1]) for xx, lq in zip(x, log_q)]
    for actual_call, expected_call in zip(
        proposal.compute_weights.call_args_list,
        compute_weights_calls,
    ):
        assert_structured_arrays_equal(actual_call[0][0], expected_call[0])
        np.testing.assert_array_equal(actual_call[0][1], expected_call[1])



# -*- coding: utf-8 -*-
"""
Test the FlowModel object.
"""
import json
import numpy as np
import os
import pickle
import pytest
import torch
from unittest.mock import create_autospec, MagicMock, Mock, patch

import torch.utils

from nessai.flowmodel import FlowModel
from nessai.flowmodel import config
from nessai.flows.realnvp import RealNVP


@pytest.fixture()
def data_dim():
    return 2


@pytest.fixture()
def n_samples():
    return 10


@pytest.fixture(params=[False, True])
def conditional(request, n_samples):
    if request.param:
        return np.random.randn(n_samples, 2)
    else:
        return None


@pytest.fixture()
def model():
    m = create_autospec(FlowModel)
    m.numpy_array_to_tensor = torch.tensor
    m.model = MagicMock()
    return m


@pytest.fixture(scope="function")
def flow_model(flow_config, data_dim, tmpdir):
    flow_config["model_config"]["n_inputs"] = data_dim
    output = str(tmpdir.mkdir("flowmodel"))
    return FlowModel(flow_config, output=output)


def test_init_no_config(tmpdir):
    """Test the init method with no config specified"""
    output = str(tmpdir.mkdir("no_config"))
    default_config = config.DEFAULT_FLOW_CONFIG.copy()
    default_config["model_config"] = config.DEFAULT_MODEL_CONFIG.copy()
    fm = FlowModel(output=output)

    assert fm.model_config == default_config["model_config"]


def test_init_no_output(model, tmpdir):
    """Assert the current working directory is used by default"""
    output = str(tmpdir.mkdir("default_output"))
    with patch("os.getcwd", return_value=output) as mock:
        FlowModel.__init__(model, output=None)
    mock.assert_called_once()
    assert model.output is output


def test_init_config_class(tmpdir):
    """Test the init and save methods when specifying `flow` as a class"""

    output = str(tmpdir.mkdir("no_config"))
    config = dict(model_config=dict(flow=RealNVP))
    fm = FlowModel(config=config, output=output)

    assert fm.model_config["flow"].__name__ == "RealNVP"


def test_save_input(model, tmp_path):
    """Test the save input function"""
    output = tmp_path / "test"
    output.mkdir()
    model.output = output

    config = dict(
        patience=10,
        x=np.array([1, 2]),
        model_config=dict(
            n_neurons=10,
            mask=np.array([1, 0]),
            flow=RealNVP,
        ),
    )

    FlowModel.save_input(model, config, output_file=None)

    file_path = os.path.join(output, "flow_config.json")
    assert os.path.exists(file_path)
    with open(file_path, "r") as fp:
        d = json.load(fp)
    assert d["x"] == "[1,2]"
    assert d["model_config"]["mask"] == "[1,0]"


def test_initialise(model):
    """Test the initialise method"""
    model.get_optimiser = MagicMock()
    model.model_config = dict(n_neurons=2)
    model.inference_device = None
    model.optimiser = "adam"
    model.optimiser_kwargs = {"weights": 0.1}
    with patch(
        "nessai.flowmodel.base.configure_model", return_value=("model", "cpu")
    ) as mock:
        FlowModel.initialise(model)
    mock.assert_called_once_with(model.model_config)
    model.get_optimiser.assert_called_once_with("adam", weights=0.1)
    assert model.inference_device == torch.device("cpu")


@pytest.mark.parametrize("optimiser", ["Adam", "AdamW", "SGD"])
def test_get_optimiser(model, optimiser):
    """Test to make sure the three supported optimisers work"""
    model.lr = 0.1
    model.model = MagicMock()
    with patch(f"torch.optim.{optimiser}") as mock:
        FlowModel.get_optimiser(model, optimiser)
    mock.assert_called_once()


@pytest.mark.parametrize(
    "data_size, batch_size",
    [(4010, 1000), (106, 21), (1000, 1000), (2000, 1000)],
)
def test_check_batch_size(data_size, batch_size):
    """"""
    x = np.arange(data_size)
    out = FlowModel.check_batch_size(x, batch_size)
    assert out >= int(0.1 * batch_size)
    if not data_size % batch_size:
        assert out == batch_size


def test_check_batch_size_min_size():
    """Make sure the minimum batch size is respected so long as the final batch
    has size > 1.

    In this case the minimum valid size is 8 but that results in the final
    batch having a size of 1, so it should be 7 instead.
    """
    x = np.arange(25)
    out = FlowModel.check_batch_size(x, 10, min_fraction=0.8)
    assert out == 7


def test_check_batch_size_2():
    """Assert an error is raised if the batch size reaches less than two"""
    x = np.arange(3)
    with pytest.raises(RuntimeError) as excinfo:
        FlowModel.check_batch_size(x, 2, min_fraction=1.0)
    assert "Could not find a valid batch size" in str(excinfo.value)


def test_check_batch_size_1():
    """Assert an error is raised if the batch size is 1"""
    x = np.arange(2)
    with pytest.raises(ValueError) as excinfo:
        FlowModel.check_batch_size(x, 1, min_fraction=1.0)
    assert "Cannot use a batch size of 1" in str(excinfo.value)


@pytest.mark.parametrize("val_size, batch_size", [(0.1, 100), (0.5, "all")])
def test_prep_data(flow_model, data_dim, val_size, batch_size):
    """
    Test the data prep, make sure batch sizes and validation size
    produce the correct result.
    """
    n = 100
    x = np.random.randn(n, data_dim)

    train, val, batch_size_out = flow_model.prep_data(x, val_size, batch_size)
    if batch_size == "all":
        batch_size = int(n * (1 - val_size))

    assert flow_model._batch_size == batch_size
    assert len(train) + len(val) == n


@pytest.mark.parametrize(
    "val_size, batch_size", [(0.1, 100), (0.5, "all"), (0.1, None)]
)
def test_prep_data_dataloader(flow_model, data_dim, val_size, batch_size):
    """
    Test the data prep, make sure batch sizes and validation size
    produce the correct result.
    """
    n = 100
    x = np.random.randn(n, data_dim)

    train, val, batch_size_out = flow_model.prep_data(
        x, val_size, batch_size, use_dataloader=True
    )
    train_batch = next(iter(train))[0]
    val_batch = next(iter(val))[0]
    if batch_size == "all" or batch_size is None:
        batch_size = int(n * (1 - val_size))

    assert train.batch_size == batch_size
    assert list(val_batch.shape) == [int(val_size * n), data_dim]
    assert len(train) * train_batch.shape[0] + val_batch.shape[0] == n


def test_prep_data_conditional(data_dim):
    n = 200
    batch_size = 100
    x = np.random.randn(n, data_dim)
    c = np.random.randn(n, 1)
    fm = create_autospec(FlowModel)
    fm.initialised = True
    fm.check_batch_size = MagicMock(return_value=batch_size)
    train_loader, val_loader, bs = FlowModel.prep_data(
        fm, x, 0.1, batch_size, conditional=c
    )
    assert bs == batch_size
    assert fm._batch_size == batch_size
    batch = next(iter(train_loader))
    assert len(batch) == 2
    assert batch[0].shape == (batch_size, data_dim)
    assert batch[1].shape == (batch_size, 1)
    batch = next(iter(val_loader))
    assert len(batch) == 2
    assert batch[0].shape == (20, data_dim)
    assert batch[1].shape == (20, 1)


@pytest.mark.parametrize("batch_size", ["10", True, False])
def test_incorrect_batch_size_type(flow_model, data_dim, batch_size):
    """Ensure the non-interger batch sizes do not work"""
    n = 1000
    x = np.random.randn(n, data_dim)
    with pytest.raises(RuntimeError) as excinfo:
        flow_model.prep_data(x, 0.1, batch_size)
    assert "Unknown batch size" in str(excinfo.value)


@pytest.mark.parametrize(
    "x", [np.array([np.inf]), np.array([-np.inf]), np.array([np.nan])]
)
def test_prep_data_non_finite_values(model, x):
    """Assert an error is raised if the samples contain non-finite values"""
    model.initialised = True
    with pytest.raises(
        ValueError, match=r"Cannot train with non-finite samples!"
    ):
        FlowModel.prep_data(model, x, 0.1, 10)


@pytest.mark.parametrize(
    "w", [np.array([np.inf]), np.array([-np.inf]), np.array([np.nan])]
)
def test_prep_data_non_finite_weights(model, w):
    """Assert an error is raised if the weights contain non-finite values"""
    model.initialised = True
    with pytest.raises(ValueError, match=r"Weights contain non-finite values"):
        FlowModel.prep_data(model, np.random.rand(100), 0.1, 50, weights=w)


def test_prep_data_weights_and_conditional(model):
    """Assert an error is raised if weights and conditional are specified"""
    model.initialised = True
    with pytest.raises(
        RuntimeError, match=r"weights and conditional inputs not supported"
    ):
        FlowModel.prep_data(
            model,
            np.random.rand(100),
            0.1,
            50,
            weights=np.random.rand(100),
            conditional=np.random.rand(100),
        )


@pytest.mark.parametrize("dataloader", [False, True])
def test_training(flow_model, data_dim, dataloader):
    """Test class until training"""
    x = np.random.randn(200, data_dim)
    flow_model.use_dataloader = dataloader
    flow_model.train(x)
    assert flow_model.weights_file is not None


def test_training_with_weights(flow_model, data_dim):
    """Test training with weights"""
    x = np.random.randn(200, data_dim)
    weights = np.random.rand(200)
    flow_model.train(x, weights=weights)
    assert flow_model.weights_file is not None


def test_training_with_conditional(data_dim, tmp_path):
    """Test training with conditional inputs"""
    output = tmp_path / "test_train_conditional"
    fm = create_autospec(FlowModel)
    fm.batch_size = 100
    fm.initialised = True
    fm.noise_scale = None
    fm.device = "cpu"
    fm.inference_device = None
    fm.use_dataloader = False
    fm.annealing = False
    fm.patience = 20
    fm.prep_data = MagicMock(return_value=("train", "val", None))
    fm._train = MagicMock(return_value=1.0)
    fm._validate = MagicMock(return_value=1.0)
    x = np.random.randn(50, data_dim)
    conditional = np.random.randint(0, 2, size=(50, 1))
    FlowModel.train(
        fm,
        x,
        conditional=conditional,
        max_epochs=1,
        val_size=0.1,
        output=output,
    )
    fm.prep_data.assert_called_once_with(
        x,
        val_size=0.1,
        batch_size=100,
        weights=None,
        conditional=conditional,
        use_dataloader=True,
    )
    fm._train.assert_called_once()
    assert fm._train.call_args_list[0][1]["is_conditional"] is True
    fm._validate.assert_called_once()
    assert fm._validate.call_args_list[0][1]["is_conditional"] is True


@pytest.mark.parametrize(
    "x", [np.array([np.inf]), np.array([-np.inf]), np.array([np.nan])]
)
def test_training_non_finite_samples(model, x):
    """Assert an error is raised if the samples contain non-finite values"""
    model.initialised = True
    with pytest.raises(ValueError, match=r"Training data is not finite"):
        FlowModel.train(model, x)


@pytest.mark.parametrize(
    "kwargs",
    [
        {"annealing": True},
        {"noise_type": "constant", "noise_scale": 0.1},
        {"noise_type": "adaptive", "noise_scale": 0.1},
        {"max_epochs": 51},
        {"val_size": 0.0},
        {"val_size": None},
    ],
)
def test_training_additional_config_args(
    flow_config,
    data_dim,
    tmpdir,
    kwargs,
):
    """
    Test training with different config args
    """
    flow_config["model_config"]["n_inputs"] = data_dim
    for key, value in kwargs.items():
        flow_config[key] = value

    output = str(tmpdir.mkdir("flowmodel"))
    flow_model = FlowModel(flow_config, output=output)

    assert getattr(flow_model, key) == value

    x = np.random.randn(100, data_dim)
    flow_model.train(x)


def test_train_func_conditional(data_dim):
    n = 100

    model = Mock(spec=["train"])

    def log_prob(x, cond):
        assert len(x) == len(cond)
        return torch.randn(x.shape[0], requires_grad=True)

    model.log_prob = MagicMock(side_effect=log_prob)
    model.parameters = MagicMock(return_value=[MagicMock(), MagicMock()])

    fm = create_autospec(FlowModel)
    fm.model = model
    fm._optimiser = MagicMock(spec=torch.optim.Adam)
    fm.device = "cpu"
    fm.clip_grad_norm = 10.0
    fm.annealing = False

    data = torch.utils.data.DataLoader(
        torch.utils.data.TensorDataset(
            torch.randn(n, data_dim),
            torch.randn(n, 1),
        ),
        batch_size=n,
    )

    FlowModel._train(fm, data, is_dataloader=True, is_conditional=True)

    model.log_prob.assert_called_once()


def test_validate_func_conditional(data_dim):
    n = 100

    model = Mock(spec=["eval"])

    def log_prob(x, cond):
        assert len(x) == len(cond)
        return torch.randn(x.shape[0], requires_grad=True)

    model.log_prob = MagicMock(side_effect=log_prob)
    model.parameters = MagicMock(return_value=[MagicMock(), MagicMock()])

    fm = create_autospec(FlowModel)
    fm.model = model
    fm._optimiser = MagicMock(spec=torch.optim.Adam)
    fm.device = "cpu"
    fm.clip_grad_norm = 10.0
    fm.annealing = False

    data = torch.utils.data.DataLoader(
        torch.utils.data.TensorDataset(
            torch.randn(n, data_dim),
            torch.randn(n, 1),
        ),
        batch_size=n,
    )

    FlowModel._validate(fm, data, is_dataloader=True, is_conditional=True)

    model.log_prob.assert_called_once()


def test_early_optimiser_init(flow_model):
    """Ensure calling the opitmiser before the model raises an error"""
    with pytest.raises(RuntimeError) as excinfo:
        flow_model.get_optimiser()
    assert "Cannot initialise optimiser" in str(excinfo.value)


@pytest.mark.parametrize("weights", [False, True])
@pytest.mark.parametrize("perms", [False, True])
def test_reset_model(model, weights, perms):
    """Test resetting the model"""
    model.model = MagicMock()
    model.apply = MagicMock()
    model.get_optimiser = MagicMock()
    model.optimiser = MagicMock()
    model.optimiser_kwargs = {"lr": 0.01}

    with patch(
        "nessai.flowmodel.base.configure_model",
        return_value=(MagicMock, "cpu"),
    ) as mock:
        FlowModel.reset_model(model, weights=weights, permutations=perms)

    if weights and perms:
        mock.assert_called_once()
    if any([weights, perms]):
        model.get_optimiser.assert_called_once_with(model.optimiser, lr=0.01)
    else:
        model.get_optimiser.assert_not_called()


def test_sample_and_log_prob(flow_model, n_samples, conditional):
    """Assert the outputs have the correct shape"""
    if conditional is not None:
        flow_model.model_config["kwargs"]["context_features"] = (
            conditional.shape[-1]
        )
    flow_model.initialise()
    samples, log_prob = flow_model.sample_and_log_prob(
        n_samples, conditional=conditional
    )
    assert len(samples) == n_samples
    assert len(log_prob) == n_samples


def test_sample_and_log_prob_not_initialised(flow_model, data_dim):
    """
    Ensure user cannot call the method before the model initialise.
    """
    with pytest.raises(RuntimeError) as excinfo:
        flow_model.sample_and_log_prob()
    assert "Model is not initialised" in str(excinfo.value)


@pytest.mark.parametrize("N", [1, 100])
def test_sample_and_log_prob_no_latent(flow_model, data_dim, N):
    """
    Test the basic use of sample and log prob and ensure correct output
    shapes.
    """
    flow_model.initialise()
    x, log_prob = flow_model.sample_and_log_prob(N=N)
    assert x.shape == (N, data_dim)
    assert log_prob.size == N


@pytest.mark.parametrize("N", [1, 100])
def test_sample_and_log_prob_with_latent(flow_model, data_dim, N):
    """
    Test the basic use of sample and log prob when samples from the
    latent space are provided
    """
    flow_model.initialise()
    z = np.random.randn(N, data_dim)
    x, log_prob = flow_model.sample_and_log_prob(z=z)
    assert x.shape == (N, data_dim)
    assert log_prob.size == N


def test_numpy_array_to_tensor(model, n_samples):
    x = np.random.randn(n_samples, 2).astype("float32")
    model.model.device = "cpu"
    out = FlowModel.numpy_array_to_tensor(model, x)
    assert isinstance(out, torch.Tensor)
    assert out.dtype == torch.get_default_dtype()
    np.testing.assert_equal(x, out.numpy())


def test_sample_log_prob_alt_dist(model):
    """Assert the alternate distribution is used."""
    z = torch.randn(5, 2)
    x = torch.randn(5, 2)
    log_prob = torch.randn(5)
    log_j = torch.randn(5)
    log_prob_expected = log_prob - log_j
    model.model = MagicMock()
    model.model.device = "cpu"
    model.model.eval = MagicMock()
    model.model.base_distribution_log_prob = MagicMock()
    model.model.inverse = MagicMock(return_value=(x, log_j))
    alt_dist = MagicMock()
    alt_dist.log_prob = MagicMock(return_value=log_prob)

    x_out, log_prob_out = FlowModel.sample_and_log_prob(
        model, z=z, alt_dist=alt_dist
    )

    model.model.inverse.assert_called_once_with(z, context=None)
    model.model.base_distribution_log_prob.assert_not_called()
    alt_dist.log_prob.assert_called_once_with(z)
    np.testing.assert_equal(x_out, x.numpy())
    np.testing.assert_equal(log_prob_out, log_prob_expected)


def test_forward_and_log_prob(model, n_samples, conditional):
    """Assert the method from the flow is called"""
    x = np.random.randn(n_samples, 2)
    log_prob = torch.randn(n_samples)
    z = torch.randn(n_samples, 2)
    model.model = MagicMock()
    model.model.device = "cpu"
    model.model.eval = MagicMock()
    model.model.forward_and_log_prob = MagicMock(return_value=(z, log_prob))

    out_z, out_log_prob = FlowModel.forward_and_log_prob(
        model, x, conditional=conditional
    )

    model.model.eval.assert_called_once()
    model.model.forward_and_log_prob.assert_called_once()
    if conditional is not None:
        assert np.array_equal(
            model.model.forward_and_log_prob.call_args_list[0][1]["context"],
            conditional,
        )
    np.testing.assert_equal(out_z, z.numpy())
    np.testing.assert_equal(out_log_prob, log_prob.numpy())


def test_log_prob(model, n_samples, conditional):
    """Assert the correct method from the flow is called"""
    x = np.random.randn(n_samples, 2)
    log_prob = torch.randn(n_samples)
    model.model = MagicMock()
    model.model.device = "cpu"
    model.model.eval = MagicMock()
    model.model.log_prob = MagicMock(return_value=log_prob)

    out = FlowModel.log_prob(model, x, conditional=conditional)

    model.model.eval.assert_called_once()
    if conditional is not None:
        assert np.array_equal(
            model.model.log_prob.call_args_list[0][1]["context"], conditional
        )
    np.testing.assert_equal(out, log_prob.numpy())


def test_sample(model, n_samples, conditional):
    """Assert the correct method from the flow is called."""
    x = torch.randn(n_samples, 2)
    model.model = MagicMock()
    model.model.sample = MagicMock(return_value=x)

    out = FlowModel.sample(model, n_samples, conditional=conditional)

    model.model.sample.assert_called_once()
    assert model.model.sample.call_args_list[0][0][0] == n_samples
    if conditional is not None:
        assert np.array_equal(
            model.model.sample.call_args_list[0][1]["context"], conditional
        )
    np.testing.assert_array_equal(out, x.numpy())


def test_sample_latent_distribution(model):
    """Assert the correct method is called"""
    n = 10
    z = torch.randn(n, 2)
    model.model = MagicMock()
    model.model.sample_latent_distribution = MagicMock(return_value=z)
    out = FlowModel.sample_latent_distribution(model, n)
    model.model.sample_latent_distribution.assert_called_once_with(n)
    np.testing.assert_array_equal(out, z.numpy())


def test_move_to_update_default(model):
    """Ensure the stored device is updated"""
    model.device = "cuda"
    model.model = MagicMock()
    model.model.to = MagicMock()
    FlowModel.move_to(model, "cpu", update_default=True)
    assert model.device == torch.device("cpu")
    model.model.to.assert_called_once()


@patch("torch.save")
def test_save_weights(mock_save, model):
    """Test saving the weights"""
    model.model = MagicMock()
    FlowModel.save_weights(model, "model.pt")
    mock_save.assert_called_once()
    assert model.weights_file == "model.pt"


@patch("torch.save")
@patch("shutil.move")
@patch("os.path.exists", return_value=True)
def test_save_weights_existing(mock_save, mock_move, mock_exists, model):
    """Assert the file is move to a file with the correct name."""
    model.model = MagicMock()
    FlowModel.save_weights(model, "model.pt")
    mock_save.assert_called_once()
    mock_move.assert_called_once_with("model.pt", "model.pt.old")
    assert model.weights_file == "model.pt"


@pytest.mark.parametrize("initialised", [False, True])
def test_load_weights(model, initialised):
    """Assert the correct functions are called with the correct inputs"""
    weights_file = "test.pt"
    model.initialised = initialised
    model.initialise = MagicMock()
    model.model = MagicMock()
    model.model.load_state_dict = MagicMock()
    model.model.eval = MagicMock()
    d = dict(weight=torch.tensor(1))
    with patch("torch.load", return_value=d) as mock_load:
        FlowModel.load_weights(model, weights_file)
    # Shouldn't initialise twice
    if initialised:
        model.initialise.assert_not_called()
    else:
        model.initialise.assert_called_once()
    mock_load.assert_called_once_with(weights_file)
    model.model.load_state_dict.assert_called_once_with(d)
    model.model.eval.assert_called_once()
    assert model.weights_file == weights_file


def test_reload_weights(model):
    """Assert the correct weights file is used."""
    model.load_weights = MagicMock()
    model.weights_file = "test.pt"
    FlowModel.reload_weights(model, None)
    model.load_weights.assert_called_once_with("test.pt")


def test_get_state(flow_model):
    """Make the object can be pickled"""
    pickle.dumps(flow_model)


@pytest.mark.parametrize("N", [1, 100])
@pytest.mark.parametrize("conditional", [None, True])
@pytest.mark.integration_test
def test_forward_and_log_prob_integration(
    flow_model, data_dim, N, conditional
):
    """Test the basic use of forward and log prob"""
    if conditional:
        conditional = np.random.randn(N, 1)
        flow_model.model_config["kwargs"]["context_features"] = 1
    flow_model.initialise()
    x = np.random.randn(N, data_dim)
    z, log_prob = flow_model.forward_and_log_prob(x, conditional=conditional)
    assert z.shape == (N, data_dim)
    assert log_prob.size == N


@pytest.mark.integration_test
def test_lu_cache_reset(tmp_path):
    """Assert the LU cache is correctly reset after training.

    Cache is reset when calling .train() so if cache is incorrect then after
    the reset the values in the latent space will not match.
    """
    output = tmp_path / "test"
    output.mkdir()

    config = dict(
        max_epochs=100,
        patience=1000,
        model_config=dict(
            n_inputs=2,
            n_blocks=2,
            kwargs=dict(
                linear_transform="lu",
            ),
        ),
    )

    flow = FlowModel(config=config, output=output)
    data = np.random.randn(100, 2)

    flow.train(data)

    test_data = torch.from_numpy(data).type(torch.get_default_dtype())

    with torch.inference_mode():
        z_out, log_j = flow.model.forward(test_data)
    flow.model.train()
    flow.model.eval()
    with torch.inference_mode():
        z_out_reset, log_j_reset = flow.model.forward(test_data)

    np.testing.assert_array_equal(z_out_reset, z_out)
    np.testing.assert_array_equal(log_j_reset, log_j)


@pytest.mark.integration_test
def test_train_without_validation(tmp_path):
    """Assert training without validation works"""
    output = tmp_path / "test_no_validation"
    output.mkdir()

    config = dict(
        max_epochs=100,
        patience=1000,
        val_size=None,
        model_config=dict(
            n_inputs=2,
            n_blocks=2,
            kwargs=dict(
                linear_transform="lu",
            ),
        ),
    )

    flow = FlowModel(config=config, output=output)
    data = np.random.randn(100, 2)

    history = flow.train(data)

    assert np.isnan(history["val_loss"]).all()


@pytest.mark.integration_test
def test_train_conditional_integration(tmp_path):
    """Assert training with conditional data works"""
    output = tmp_path / "test_train_conditional"
    output.mkdir()

    config = dict(
        max_epochs=10,
        model_config=dict(
            n_inputs=2,
            n_blocks=2,
            kwargs=dict(
                linear_transform="lu",
                context_features=1,
            ),
        ),
    )

    flow = FlowModel(config=config, output=output)
    data = np.random.randn(100, 2)
    conditional = np.random.randint(2, size=(100, 1))

    _ = flow.train(data, conditional=conditional)



"""Tests related to the results returned by the sampler"""

import datetime
from unittest.mock import MagicMock, create_autospec

from nessai.samplers.importancesampler import (
    ImportanceNestedSampler as INS,
    OrderedSamples,
)
from nessai.evidence import _INSIntegralState
import numpy as np


def test_get_result_dictionary(ins, history, samples, iid):
    ins.training_samples.samples = samples.copy()
    ins.training_samples.state.log_evidence = -1.1
    ins.training_samples.state.log_evidence_error = 0.2
    if iid:
        ins.iid_samples = create_autospec(OrderedSamples)
        ins.iid_samples.state = create_autospec(_INSIntegralState)
        ins.iid_samples.samples = samples.copy()
        ins.iid_samples.state.log_evidence = -0.9
        ins.iid_samples.state.log_evidence_error = 0.14
    ins.final_samples = samples
    ins.history = history
    ins.state = MagicMock(spec=_INSIntegralState)

    ins.seed = 1234
    ins.sampling_time = datetime.timedelta(seconds=10)
    ins.training_time = datetime.timedelta(seconds=10)
    ins.draw_samples_time = datetime.timedelta(seconds=10)
    ins.add_and_update_samples_time = datetime.timedelta(seconds=10)
    ins.draw_final_samples_time = datetime.timedelta(seconds=10)

    ins.log_evidence = -1.1
    ins.log_evidence_error = 0.3

    ins.bootstrap_log_evidence = None
    ins.bootstrap_log_evidence_error = None

    ins.state.log_posterior_weights = np.log(np.random.rand(len(samples)))

    ins.importance = {
        "total": [1, 2, 3],
        "evidence": [1, 2, 3],
        "posterior": [1, 2, 3],
    }

    out = INS.get_result_dictionary(ins)

    assert isinstance(out, dict)



"""
Test the ImportanceFlowModel.
"""

import os
import pickle
from unittest.mock import MagicMock, create_autospec, patch

from nessai.flowmodel.importance import ImportanceFlowModel as IFM
import numpy as np
import pytest
import torch


@pytest.fixture()
def ifm():
    return create_autospec(IFM)


class DummyFlow(torch.nn.Module):
    pass


def test_init(ifm):
    config = dict(patience=20)
    output = "test"
    with patch("nessai.flowmodel.importance.FlowModel.__init__") as mock_init:
        IFM.__init__(ifm, output=output, config=config)
    mock_init.assert_called_once_with(config=config, output=output)
    assert ifm.weights_files == []
    assert len(ifm.models) == 0


def test_model_property_with_models(ifm):
    """Assert last model is returned"""
    models = [DummyFlow(), DummyFlow(), DummyFlow()]
    ifm.models = models
    assert IFM.model.__get__(ifm) is models[-1]


def test_model_property_without_models(ifm):
    """Assert None is return if no models have been defined"""
    ifm.models = []
    assert IFM.model.__get__(ifm) is None


def test_model_setter(ifm):
    """Assert the model setter appends the model"""
    new_model = DummyFlow()
    models = [DummyFlow(), DummyFlow()]
    ifm.models = models
    IFM.model.__set__(ifm, new_model)
    assert len(ifm.models) == 3
    assert ifm.models[-1] is new_model


def test_model_setter_none(ifm):
    """Assert none is not added to the models"""
    new_model = None
    models = [DummyFlow(), DummyFlow()]
    ifm.models = models
    IFM.model.__set__(ifm, new_model)
    assert len(ifm.models) == 2
    assert None not in ifm.models


def test_n_models(ifm):
    """Assert the number of models is correct"""
    ifm.models = [DummyFlow(), DummyFlow()]
    assert IFM.n_models.__get__(ifm) == 2


@pytest.mark.parametrize("models", [None, torch.nn.ModuleList()])
def test_n_models_no_models(ifm, models):
    """Assert the number of models is 0 if models is empty or None"""
    ifm.models = models
    assert IFM.n_models.__get__(ifm) == 0


def test_initialise(ifm):
    """Assert initialise sets the correct boolean"""
    ifm.initialised = False
    IFM.initialise(ifm)
    assert ifm.initialised is True


def test_reset_optimiser(ifm):
    """Assert the optimiser is reset"""
    optimiser = "adam"
    optimiser_kwargs = dict(decay=0.1)
    _optimiser = object()
    ifm.get_optimiser = MagicMock(return_value=_optimiser)
    ifm.optimiser = optimiser
    ifm.optimiser_kwargs = optimiser_kwargs
    IFM.reset_optimiser(ifm)
    ifm.get_optimiser.assert_called_once_with(
        optimiser=optimiser,
        **optimiser_kwargs,
    )
    assert ifm._optimiser is _optimiser


def test_add_new_flow_reset(ifm):
    """Assert a new flow is created when reset=True."""
    flow = DummyFlow()
    device = torch.device("cpu")

    ifm.models = torch.nn.ModuleList([DummyFlow(), DummyFlow()])
    ifm.model_config = dict(inference_device_tag=None)
    ifm.reset_optimiser = MagicMock()

    with patch(
        "nessai.flowmodel.importance.configure_model",
        return_value=(flow, device),
    ) as mock_configure:
        IFM.add_new_flow(ifm, reset=True)

    assert len(ifm.models) == 3
    assert ifm.models[-1] is flow
    assert ifm.models.training is False

    mock_configure.assert_called_once_with(ifm.model_config)

    ifm.reset_optimiser.assert_called_once()
    assert ifm.device == device
    assert ifm.inference_device == device


def test_add_new_flow_no_reset(ifm):
    """Assert the flow is copied when reset=False"""
    flow = DummyFlow()
    device = torch.device("cpu")

    flow_to_copy = DummyFlow()

    ifm.device = device
    ifm.model = flow_to_copy
    ifm.models = torch.nn.ModuleList([DummyFlow(), flow_to_copy])
    ifm.model_config = dict(inference_device_tag=None)
    ifm.reset_optimiser = MagicMock()

    with patch(
        "copy.deepcopy",
        return_value=flow,
    ) as mock_copy:
        IFM.add_new_flow(ifm, reset=False)

    assert len(ifm.models) == 3
    assert ifm.models[-1] is flow
    assert ifm.models.training is False

    mock_copy.assert_called_once_with(ifm.models[-2])

    ifm.reset_optimiser.assert_called_once()
    assert ifm.device == device
    assert ifm.inference_device == device


def test_add_new_flow_first_flow(ifm):
    """Assert a new flow is created when no flows have been added"""
    flow = DummyFlow()
    device = torch.device("cpu")

    ifm.models = torch.nn.ModuleList()
    ifm.model_config = dict(inference_device_tag=None)
    ifm.reset_optimiser = MagicMock()

    with patch(
        "nessai.flowmodel.importance.configure_model",
        return_value=(flow, device),
    ) as mock_configure:
        IFM.add_new_flow(ifm, reset=False)

    assert len(ifm.models) == 1
    assert ifm.models[-1] is flow
    assert ifm.models.training is False

    mock_configure.assert_called_once_with(ifm.model_config)

    ifm.reset_optimiser.assert_called_once()
    assert ifm.device == device
    assert ifm.inference_device == device


def test_add_new_flow_first_inference_device(ifm):
    """Assert the inference device is set correctly"""
    flow = DummyFlow()
    device = torch.device("cpu")

    ifm.models = torch.nn.ModuleList()
    ifm.model_config = dict(inference_device_tag="cuda")

    dummy_device = object()

    with patch(
        "nessai.flowmodel.importance.configure_model",
        return_value=(flow, device),
    ), patch("torch.device", return_value=dummy_device) as mock_device:
        IFM.add_new_flow(ifm, reset=False)

    mock_device.assert_called_once_with("cuda")
    assert ifm.inference_device is dummy_device


@pytest.mark.parametrize("i", [0, 1, 2, -1])
def test_log_prob_th(ifm, i):
    """Assert correct values are returned"""
    x = np.random.randn(10, 2)
    log_prob = torch.ones(len(x), dtype=torch.float32)

    model = MagicMock(spec=DummyFlow)
    model.log_prob = MagicMock(return_value=log_prob)
    model.device = torch.device("cpu")
    model.training = True
    model.eval = MagicMock()

    models = [DummyFlow(), DummyFlow(), DummyFlow()]
    models[i] = model
    ifm.models = torch.nn.ModuleList(models)
    ifm.models.training = True

    out = IFM.log_prob_ith(ifm, x, i)

    assert out.dtype == np.float64
    model.eval.assert_called_once()
    np.testing.assert_array_equal(out, log_prob.numpy())


def test_log_prob_all(ifm):
    x = np.random.randn(10, 2)
    log_prob = [
        torch.ones(len(x), dtype=torch.float32),
        2 * torch.ones(len(x), dtype=torch.float32),
        3 * torch.ones(len(x), dtype=torch.float32),
    ]

    models = [
        MagicMock(spec=DummyFlow, log_prob=MagicMock(return_value=lp))
        for lp in log_prob
    ]

    n_expected = len(models)

    ifm.models = torch.nn.ModuleList(models)
    ifm.models.training = True
    ifm.model.device = torch.device("cpu")
    ifm.n_models = len(models)

    out = IFM.log_prob_all(ifm, x)

    assert ifm.models.training is False
    assert out.dtype == np.float64
    assert out.shape == (10, n_expected)

    for model in models[:n_expected]:
        model.log_prob.assert_called_once()


@pytest.mark.parametrize("i", [0, 1, 2, -1])
def test_sample_ith(ifm, i):
    """Assert the correct number of samples is drawn from the correct model"""
    n = 10
    x = torch.randn(n, 2)

    model = MagicMock(spec=DummyFlow)
    model.sample = MagicMock(return_value=x)
    model.device = torch.device("cpu")
    model.training = True
    model.eval = MagicMock()

    models = [DummyFlow(), DummyFlow(), DummyFlow()]
    models[i] = model
    ifm.models = torch.nn.ModuleList(models)
    ifm.models.training = True

    out = IFM.sample_ith(ifm, i, N=n)

    assert out.dtype == np.float64
    model.eval.assert_called_once()
    model.sample.assert_called_once_with(n)
    np.testing.assert_array_equal(out, x.numpy())


def test_sample_ith_error(ifm):
    """Assert an error is raised if no models have been added"""
    ifm.models = None
    with pytest.raises(RuntimeError, match=r"Models are not initialised yet!"):
        IFM.sample_ith(ifm, 1)


def test_save_weights(ifm):
    """Assert the correct method is called and the weights file is stored"""
    file = "w1.pt"
    ifm.weights_files = ["w0.pt"]

    def func(f):
        ifm.weights_file = f

    with patch(
        "nessai.flowmodel.importance.FlowModel.save_weights", side_effect=func
    ) as mock_save:
        IFM.save_weights(ifm, file)
    mock_save.assert_called_once_with(file)
    assert ifm.weights_files == ["w0.pt", file]


def test_load_all_weights(ifm):
    """Assert all the weights files are loaded"""
    weights_files = ["w0.pt", "w1.pt", "w2.pt"]
    weights = ["w0", "w1", "w2"]
    device = torch.device("cpu")
    ifm.weights_files = weights_files
    models = torch.nn.ModuleList(
        [
            MagicMock(spec=DummyFlow(), load_state_dict=MagicMock())
            for _ in range(len(weights_files))
        ]
    )
    config_out = [(f, device) for f in models]
    with patch(
        "nessai.flowmodel.importance.configure_model",
        side_effect=config_out,
    ) as mock_configure, patch("torch.load", side_effect=weights):
        IFM.load_all_weights(ifm)

    assert len(mock_configure.call_args_list) == 3
    assert all(m is n for m, n in zip(models, ifm.models))
    assert ifm.models.training is False

    for model, w in zip(models, weights):
        model.load_state_dict.assert_called_once_with(w)


@pytest.mark.parametrize("n", [None, 10, 16])
def test_update_weights_path(ifm, tmp_path, n):
    """Assert the list of weights files is correctly updated"""
    path = tmp_path / "outdir"
    path.mkdir()
    n_models = 15
    n_total = 16
    expected_files = []
    for i in range(n_total):
        d = path / f"level_{i}"
        d.mkdir()
        file = d / "model.pt"
        file.write_text("data")
        file = str(file)
        expected_files.append(file)

    ifm.n_models = n_models
    IFM.update_weights_path(ifm, str(path), n=n)
    n_expeceted = n or n_models
    assert ifm.weights_files == expected_files[:n_expeceted]


def test_update_weights_path_cannot_update(ifm):
    """Assert the list of weights files is correctly updated"""
    ifm.n_models = 0
    with pytest.raises(RuntimeError, match=r"n is None and .*"):
        IFM.update_weights_path(ifm, ".", n=None)


def test_update_weights_path_not_enough_files(ifm, tmp_path):
    """Assert the list of weights files is correctly updated"""
    path = tmp_path / "outdir"
    path.mkdir()
    n_models = 5
    n_total = 4
    expected_files = []
    for i in range(n_total):
        d = path / f"level_{i}"
        d.mkdir()
        file = d / "model.pt"
        file.write_text("data")
        file = str(file)
        expected_files.append(file)

    ifm.n_models = n_models
    with pytest.raises(RuntimeError, match=r".* Not enough files."):
        IFM.update_weights_path(ifm, str(path))


@pytest.mark.parametrize("weights_path", [None, "weights_directory"])
def test_resume(ifm, weights_path):
    """Assert resume method calls the correct methods"""
    model_config = dict(patience=20)
    updated_config = dict(patience=20, ftype="realnvp")
    output = "current_output"

    expected_path = weights_path or output

    ifm.output = output
    ifm.update_weights_path = MagicMock()
    ifm.load_all_weights = MagicMock()
    ifm.initialise = MagicMock()
    ifm._resume_n_models = 10

    with patch(
        "nessai.flowmodel.importance.update_model_config",
        return_value=updated_config,
    ) as mock_update:
        IFM.resume(ifm, model_config, weights_path)

    mock_update.assert_called_once_with(model_config)

    ifm.update_weights_path.assert_called_once_with(
        expected_path,
        n=10,
    )
    ifm.load_all_weights.assert_called_once_with
    ifm.initialise.assert_called_once()


def test_getstate(ifm):
    """Assert the correct keys are added and removed"""
    ifm.models = torch.nn.ModuleList([DummyFlow(), DummyFlow()])
    ifm._optimiser = MagicMock()
    ifm.model_config = dict(patience=20)
    ifm.initialised = True
    ifm.test = "value"

    out = IFM.__getstate__(ifm)

    assert "_optimiser" not in out
    assert "model_config" not in out
    assert out["models"] is None
    assert out["_resume_n_models"] == 2
    assert out["initialised"] is False
    assert out["test"] == "value"


@pytest.mark.integration_test
def test_resume_integration(tmp_path):
    output = tmp_path / "test"
    output.mkdir()

    config = dict(model_config=dict(n_inputs=2))

    ifm = IFM(output=output, config=config)

    ifm.initialise()

    n_models = 3

    for i in range(n_models):
        level_dir = output / f"level_{i}"
        level_dir.mkdir()
        ifm.add_new_flow()
        ifm.save_weights(os.path.join(level_dir, "model.pt"))

    pickled_ifm = pickle.dumps(ifm)

    resumed_ifm = pickle.loads(pickled_ifm)

    resumed_ifm.resume(config["model_config"])

    assert len(resumed_ifm.models) == n_models



# -*- coding: utf-8 -*-
"""
Utilities for configuring torch.
"""
import logging

import torch

logger = logging.getLogger(__name__)

dtype_mapping = {
    "float32": torch.float32,
    "float64": torch.float64,
}


def set_torch_default_dtype(dtype) -> torch.dtype:
    """Set the default dtype for torch tenors.

    If dtype is None, returns the default dtype.

    Parameters
    ----------
    dtype : {"float32", "float64", torch.dtype}
        The new default dtype for torch.

    Returns
    -------
    torch.dtype
        The torch dtype used to set the default.
    """
    if dtype is None:
        return torch.get_default_dtype()
    if isinstance(dtype, str):
        if dtype not in dtype_mapping:
            raise ValueError(
                f"Unknown torch dtype: {dtype}. "
                f"Choose from: {list(dtype_mapping.keys())}"
            )
        dtype = dtype_mapping.get(dtype)
    logger.info(f"Setting torch dtype to {dtype}")
    torch.set_default_dtype(dtype)
    return dtype



# -*- coding: utf-8 -*-
"""
Tests for utilities related to producing histograms.
"""
import numpy as np
import pytest

from nessai.utils.hist import auto_bins


def test_auto_bins_max_bins():
    """Test the autobin function returns the max bins"""
    assert auto_bins(np.random.rand(100), max_bins=2) <= 2


def test_auto_bins_single_point():
    """Test to ensure the function produces a result with one sample"""
    assert auto_bins(np.random.rand()) >= 1


def test_auto_bins_no_samples():
    """Test to ensure the function produces a result with one sample"""
    with pytest.raises(RuntimeError) as excinfo:
        assert auto_bins([])
    assert "Input array is empty!" in str(excinfo.value)



# -*- coding: utf-8 -*-
"""
FlowModel for use in importance nested sampling.
"""
import copy
import glob
import logging
import os
from typing import Optional

import numpy as np
import torch

from .base import FlowModel
from .utils import update_model_config

from ..flows import configure_model


logger = logging.getLogger(__name__)


class ImportanceFlowModel(FlowModel):
    """Flow Model that contains multiple flows for importance sampler."""

    models: torch.nn.ModuleList = None
    _resume_n_models: int = None

    def __init__(self, config=None, output=None):
        super().__init__(config=config, output=output)
        self.weights_files = []
        self.models = torch.nn.ModuleList()

    @property
    def model(self):
        """The current flow (model).

        Returns None if the no models have been added.
        """
        if self.models:
            return self.models[-1]
        else:
            logger.warning("Model not defined yet!")
            return None

    @model.setter
    def model(self, model):
        if model is not None:
            self.models.append(model)

    @property
    def n_models(self) -> int:
        """Number of models (flows)"""
        if self.models:
            return len(self.models)
        else:
            return 0

    def initialise(self) -> None:
        """Initialise things"""
        self.initialised = True

    def reset_optimiser(self) -> None:
        """Reset the optimiser to point at current model.

        Uses the original optimiser and kwargs.
        """
        self._optimiser = self.get_optimiser(
            optimiser=self.optimiser, **self.optimiser_kwargs
        )

    def add_new_flow(self, reset=False):
        """Add a new flow"""
        logger.debug("Add a new flow")
        if reset or not self.models:
            new_flow, self.device = configure_model(self.model_config)
        else:
            new_flow = copy.deepcopy(self.model)
        logger.debug(f"Training device: {self.device}")
        self.inference_device = torch.device(
            self.model_config.get("inference_device_tag", self.device)
            or self.device
        )
        logger.debug(f"Inference device: {self.inference_device}")
        self.models.eval()
        self.models.append(new_flow)
        self.reset_optimiser()

    def log_prob_ith(self, x, i):
        """Compute the log-prob for the ith flow"""
        x = (
            torch.from_numpy(x)
            .type(torch.get_default_dtype())
            .to(self.models[i].device)
        )
        if self.models[i].training:
            self.models[i].eval()
        with torch.no_grad():
            log_prob = self.models[i].log_prob(x)
        log_prob = log_prob.cpu().numpy().astype(np.float64)
        return log_prob

    def log_prob_all(self, x):
        """Compute the log probability using all of the stored models."""
        x = (
            torch.from_numpy(x)
            .type(torch.get_default_dtype())
            .to(self.model.device)
        )
        if self.models.training:
            self.models.eval()
        n = self.n_models
        log_prob = torch.empty(x.shape[0], n)
        with torch.no_grad():
            for i, m in enumerate(self.models[:n]):
                log_prob[:, i] = m.log_prob(x)
        log_prob = log_prob.cpu().numpy().astype(np.float64)
        return log_prob

    def sample_ith(self, i, N=1):
        """Draw samples from the ith flow"""
        if self.models is None:
            raise RuntimeError("Models are not initialised yet!")
        if self.models[i].training:
            self.models[i].eval()

        with torch.no_grad():
            x = self.models[i].sample(int(N))

        x = x.cpu().numpy().astype(np.float64)
        return x

    def save_weights(self, weights_file) -> None:
        """Save the weights file."""
        super().save_weights(weights_file)
        self.weights_files.append(self.weights_file)

    def load_all_weights(self) -> None:
        """Load all of the weights files for each flow.

        Resets any existing models.
        """
        self.models = torch.nn.ModuleList()
        logger.debug(f"Loading weights from {self.weights_files}")
        for wf in self.weights_files:
            new_flow, self.device = configure_model(self.model_config)
            new_flow.load_state_dict(torch.load(wf))
            self.models.append(new_flow)
        self.models.eval()

    def update_weights_path(
        self, weights_path: str, n: Optional[int] = None
    ) -> None:
        """Update the weights path.

        Searches in the specified directory for weights files.

        Parameters
        ----------
        weights_path : str
            Path to the directory that contains the weights files.
        n : Optional[int]
            The number of files to load. If not specified, :code:`n_models` is
            used instead. Must be specified when resuming since the models list
            is not saved.
        """
        all_weights_files = glob.glob(
            os.path.join(weights_path, "", "level_*", "model.pt")
        )

        if n is None:
            if self.n_models:
                n = self.n_models
            else:
                raise RuntimeError(
                    "n is None and no models are defined, cannot update "
                    "weights path."
                )

        logger.debug(f"Loading weights from: {all_weights_files}")
        if len(all_weights_files) < n:
            raise RuntimeError(
                f"Cannot use weights from: {weights_path}. Not enough files."
            )
        elif len(all_weights_files) > n:
            logger.warning(
                "More weights files than expected. Some files will be skipped."
            )
        self.weights_files = [
            os.path.join(weights_path, f"level_{i}", "model.pt")
            for i in range(n)
        ]

    def resume(
        self, model_config: dict, weights_path: Optional[str] = None
    ) -> None:
        """Resume the model"""
        self.model_config = update_model_config(model_config)
        if weights_path is None:
            logger.debug(
                "Not weights path specified, looking in output directory"
            )
            weights_path = self.output
        self.update_weights_path(weights_path, n=self._resume_n_models)
        self.load_all_weights()
        self.initialise()

    def __getstate__(self):
        d = self.__dict__
        # Avoid making a copy because models can be large and this doubles the
        # memory usage.
        exclude = {"models", "_optimiser", "model_config"}
        state = {k: d[k] for k in d.keys() - exclude}
        state["initialised"] = False
        state["models"] = None
        state["_resume_n_models"] = len(d["models"])
        return state
